# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys

import pytest
import numpy
numba = pytest.importorskip("numba")

import awkward1

py27 = 2 if sys.version_info[0] < 3 else 1

def test_numpyarray_boxing():
    a = numpy.arange(10)
    wrapped = awkward1.layout.NumpyArray(a)
    assert (sys.getrefcount(a), sys.getrefcount(wrapped)) == (3, 2)

    @numba.njit
    def f1(q):
        pass
    f1(wrapped)

    @numba.njit
    def f2(q):
        return q
    out = f2(wrapped)

    assert (sys.getrefcount(a), sys.getrefcount(wrapped)) == (3, 2 + 1*py27)
    assert numpy.asarray(out).tolist() == list(range(10))
    del out
    assert (sys.getrefcount(a), sys.getrefcount(wrapped)) == (3, 2)

def test_numpyarray_refcount1():
    a = numpy.arange(10)
    wrapped = awkward1.layout.NumpyArray(a)
    assert (sys.getrefcount(a), sys.getrefcount(wrapped)) == (3, 2)
    @numba.njit
    def f1(q):
        return q[1:9][1:8][1:7]
    out = f1(wrapped)
    assert (sys.getrefcount(a), sys.getrefcount(wrapped)) == (3, 2 + 1*py27)
    assert numpy.asarray(out).tolist() == list(range(10)[1:9][1:8][1:7])
    del out
    assert (sys.getrefcount(a), sys.getrefcount(wrapped)) == (3, 2)

def test_numpyarray_refcount2():
    a = numpy.arange(10)
    wrapped = awkward1.layout.NumpyArray(a)
    assert (sys.getrefcount(a), sys.getrefcount(wrapped)) == (3, 2)
    @numba.njit
    def f1(q):
        return q[1:9], q[2:8], q[3:7]
    out = f1(wrapped)
    assert (sys.getrefcount(a), sys.getrefcount(wrapped)) == (3, 2 + 1*py27)
    assert numpy.asarray(out[-1]).tolist() == list(range(3, 7))
    del out
    assert (sys.getrefcount(a), sys.getrefcount(wrapped)) == (3, 2)

def test_numpyarray_len():
    a = awkward1.layout.NumpyArray(numpy.arange(10))
    @numba.njit
    def f1(q):
        return len(q)
    assert f1(a) == 10

def test_numpyarray_getitem_int():
    a = awkward1.layout.NumpyArray(numpy.arange(12).reshape(3, 4))

    @numba.njit
    def f1(q):
        return q[1]
    out = f1(a)
    assert isinstance(out, awkward1.layout.NumpyArray)
    assert numpy.asarray(out).tolist() == [4, 5, 6, 7]

    @numba.njit
    def f2(q):
        return q[1][2]
    assert f2(a) == 6

    @numba.njit
    def f3(q):
        return q[1, 2]
    assert f3(a) == 6

    @numba.njit
    def f4(q, i):
        return q[i]
    assert numpy.asarray(f4(a, 1)).tolist() == [4, 5, 6, 7]

def test_numpyarray_getitem_slice():
    a = awkward1.layout.NumpyArray(numpy.arange(12).reshape(3, 4))

    @numba.njit
    def f1(q):
        return q[1:]
    out = f1(a)
    assert isinstance(out, awkward1.layout.NumpyArray)
    assert numpy.asarray(out).tolist() == [[4, 5, 6, 7], [8, 9, 10, 11]]

    @numba.njit
    def f2(q, i):
        return q[i:]
    out = f2(a, 1)
    assert isinstance(out, awkward1.layout.NumpyArray)
    assert numpy.asarray(out).tolist() == [[4, 5, 6, 7], [8, 9, 10, 11]]

    @numba.njit
    def f3(q, i):
        return q[i]
    out = f3(a, slice(1, None))
    assert isinstance(out, awkward1.layout.NumpyArray)
    assert numpy.asarray(out).tolist() == [[4, 5, 6, 7], [8, 9, 10, 11]]

def test_listoffsetarray_boxing():
    offsets = awkward1.layout.Index32(numpy.array([0, 2, 2, 3], "i4"))
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3]))
    array = awkward1.layout.ListOffsetArray32(offsets, content)
    assert (sys.getrefcount(offsets), sys.getrefcount(content), sys.getrefcount(array)) == (2, 2, 2)

    @numba.njit
    def f1(q):
        return q
    out = f1(array)
    assert (sys.getrefcount(offsets), sys.getrefcount(content), sys.getrefcount(array)) == (2, 2, 2)
    assert numpy.asarray(out.offsets).tolist() == [0, 2, 2, 3]
    assert numpy.asarray(out.content).tolist() == [1.1, 2.2, 3.3]

    array2 = awkward1.layout.ListOffsetArray32(offsets, array)
    out2 = f1(array2)
    assert (sys.getrefcount(offsets), sys.getrefcount(content), sys.getrefcount(array)) == (2, 2, 2)
    assert numpy.asarray(out2.offsets).tolist() == [0, 2, 2, 3]
    assert numpy.asarray(out2.content.offsets).tolist() == [0, 2, 2, 3]
    assert numpy.asarray(out2.content.content).tolist() == [1.1, 2.2, 3.3]

def test_listoffsetarray_len():
    offsets = awkward1.layout.Index32(numpy.array([0, 2, 2, 3], "i4"))
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3]))
    array = awkward1.layout.ListOffsetArray32(offsets, content)
    @numba.njit
    def f1(q):
        return len(q)
    assert f1(array) == 3

def test_listoffsetarray_getitem_int():
    offsets = awkward1.layout.Index32(numpy.array([0, 2, 2, 3], "i4"))
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3]))
    array = awkward1.layout.ListOffsetArray32(offsets, content)
    @numba.njit
    def f1(q):
        return q[0]
    assert numpy.asarray(f1(array)).tolist() == [1.1, 2.2]
    @numba.njit
    def f2(q, i):
        return q[i]
    assert numpy.asarray(f2(array, 0)).tolist() == [1.1, 2.2]
    assert numpy.asarray(f2(array, 1)).tolist() == []
    assert numpy.asarray(f2(array, 2)).tolist() == [3.3]

def test_listoffsetarray_getitem_slice():
    offsets = awkward1.layout.Index32(numpy.array([0, 2, 2, 3], "i4"))
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3]))
    array = awkward1.layout.ListOffsetArray32(offsets, content)
    @numba.njit
    def f1(q):
        return q[1:][1]
    assert numpy.asarray(f1(array)).tolist() == [3.3]
    @numba.njit
    def f2(q, i):
        return q[i:]
    out = f2(array, 1)
    assert numpy.asarray(out[0]).tolist() == []
    assert numpy.asarray(out[1]).tolist() == [3.3]
    @numba.njit
    def f3(q, i):
        return q[i]
    out = f3(array, slice(1, None))
    assert numpy.asarray(out[0]).tolist() == []
    assert numpy.asarray(out[1]).tolist() == [3.3]



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys

import pytest
import numpy
numba = pytest.importorskip("numba")

import awkward1
awkward1_numba_util = pytest.importorskip("awkward1._numba.util")

py27 = (sys.version_info[0] < 3)

def test_numpyarray():
    array = awkward1.layout.NumpyArray(numpy.arange(10)*1.1)

    with pytest.raises(ValueError) as excinfo:
        array[20]
    assert str(excinfo.value) == "in NumpyArray attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array[-20]
    assert str(excinfo.value) == "in NumpyArray attempting to get -20, index out of range"

    array[-20:20]

    with pytest.raises(ValueError) as excinfo:
        array[20,]
    assert str(excinfo.value) == "in NumpyArray attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array[-20,]
    assert str(excinfo.value) == "in NumpyArray attempting to get -20, index out of range"

    array[-20:20,]

    with pytest.raises(ValueError) as excinfo:
        array[2, 3]
    assert str(excinfo.value) == "in NumpyArray, too many dimensions in slice"

    with pytest.raises(ValueError) as excinfo:
        array[[5, 3, 20, 8]]
    assert str(excinfo.value) == "in NumpyArray attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array[[5, 3, -20, 8]]
    assert str(excinfo.value) == "in NumpyArray attempting to get -20, index out of range"

    array.setidentities()

    with pytest.raises(ValueError) as excinfo:
        array[20]
    assert str(excinfo.value) == "in NumpyArray attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array[-20]
    assert str(excinfo.value) == "in NumpyArray attempting to get -20, index out of range"

    array[-20:20]

    with pytest.raises(ValueError) as excinfo:
        array[20,]
    assert str(excinfo.value) == "in NumpyArray attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array[-20,]
    assert str(excinfo.value) == "in NumpyArray attempting to get -20, index out of range"

    array[-20:20,]

    with pytest.raises(ValueError) as excinfo:
        array[2, 3]
    assert str(excinfo.value) == "in NumpyArray, too many dimensions in slice"

    with pytest.raises(ValueError) as excinfo:
        array[[5, 3, 20, 8]]
    assert str(excinfo.value) == "in NumpyArray attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array[[5, 3, -20, 8]]
    assert str(excinfo.value) == "in NumpyArray attempting to get -20, index out of range"

def test_listarray_numpyarray():
    starts  = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6]))
    stops   = awkward1.layout.Index64(numpy.array([3, 3, 5, 6]))
    content = awkward1.layout.NumpyArray(numpy.arange(10)*1.1)
    array   = awkward1.layout.ListArray64(starts, stops, content)

    with pytest.raises(ValueError) as excinfo:
        array[4]
    assert str(excinfo.value) == "in ListArray64, len(stops) < len(starts)"

    starts  = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6]))
    stops   = awkward1.layout.Index64(numpy.array([3, 3, 5, 6, 10]))
    content = awkward1.layout.NumpyArray(numpy.arange(10)*1.1)
    array   = awkward1.layout.ListArray64(starts, stops, content)

    with pytest.raises(ValueError) as excinfo:
        array[20]
    assert str(excinfo.value) == "in ListArray64 attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array[-20]
    assert str(excinfo.value) == "in ListArray64 attempting to get -20, index out of range"

    array[-20:20]

    with pytest.raises(ValueError) as excinfo:
        array[20,]
    assert str(excinfo.value) == "in RegularArray attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array[-20,]
    assert str(excinfo.value) == "in RegularArray attempting to get -20, index out of range"

    array[-20:20,]

    with pytest.raises(ValueError) as excinfo:
        array[2, 1, 0]
    assert str(excinfo.value) == "in NumpyArray, too many dimensions in slice"

    with pytest.raises(ValueError) as excinfo:
        array[[2, 0, 0, 20, 3]]
    assert str(excinfo.value) == "in RegularArray attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array[[2, 0, 0, -20, 3]]
    assert str(excinfo.value) == "in RegularArray attempting to get -20, index out of range"

    starts  = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6]))
    stops   = awkward1.layout.Index64(numpy.array([3, 3, 5, 6, 10]))
    content = awkward1.layout.NumpyArray(numpy.arange(10)*1.1)
    array   = awkward1.layout.ListArray64(starts, stops, content)

    array.setidentities()

    with pytest.raises(ValueError) as excinfo:
        array[2, 20]
    assert str(excinfo.value) == "in ListArray64 with identity [2] attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array[2, -20]
    assert str(excinfo.value) == "in ListArray64 with identity [2] attempting to get -20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array[1:][2, 20]
    assert str(excinfo.value) == "in ListArray64 with identity [3] attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array[1:][2, -20]
    assert str(excinfo.value) == "in ListArray64 with identity [3] attempting to get -20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array[2, [1, 0, 0, 20]]
    assert str(excinfo.value) == "in ListArray64 with identity [2] attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array[2, [1, 0, 0, -20]]
    assert str(excinfo.value) == "in ListArray64 with identity [2] attempting to get -20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array[1:][2, [0, 20]]
    assert str(excinfo.value) == "in ListArray64 with identity [3] attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array[1:][2, [0, -20]]
    assert str(excinfo.value) == "in ListArray64 with identity [3] attempting to get -20, index out of range"

def test_listarray_listarray_numpyarray():
    content  = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
    starts1  = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6]))
    stops1   = awkward1.layout.Index64(numpy.array([3, 3, 5, 6, 9]))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 3, 3, 5, 6, 9]))
    starts2  = awkward1.layout.Index64(numpy.array([0, 2, 3, 3]))
    stops2   = awkward1.layout.Index64(numpy.array([2, 3, 3, 5]))
    offsets2 = awkward1.layout.Index64(numpy.array([0, 2, 3, 3, 5]))

    array1 = awkward1.layout.ListArray64(starts1, stops1, content)
    array2 = awkward1.layout.ListArray64(starts2, stops2, array1)

    with pytest.raises(ValueError) as excinfo:
        array2[20]
    assert str(excinfo.value) == "in ListArray64 attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array2[20,]
    assert str(excinfo.value) == "in RegularArray attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array2[2, 20]
    assert str(excinfo.value) == "in ListArray64 attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array2[-20]
    assert str(excinfo.value) == "in ListArray64 attempting to get -20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array2[-20,]
    assert str(excinfo.value) == "in RegularArray attempting to get -20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array2[2, -20]
    assert str(excinfo.value) == "in ListArray64 attempting to get -20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array2[1, 0, 20]
    assert str(excinfo.value) == "in ListArray64 attempting to get 20, index out of range"

    array2.setidentities()

    with pytest.raises(ValueError) as excinfo:
        array2[20]
    assert str(excinfo.value) == "in ListArray64 attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array2[20,]
    assert str(excinfo.value) == "in RegularArray attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array2[2, 20]
    assert str(excinfo.value) == "in ListArray64 with identity [2] attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array2[1:][2, 20]
    assert str(excinfo.value) == "in ListArray64 with identity [3] attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array2[-20]
    assert str(excinfo.value) == "in ListArray64 attempting to get -20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array2[-20,]
    assert str(excinfo.value) == "in RegularArray attempting to get -20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array2[2, -20]
    assert str(excinfo.value) == "in ListArray64 with identity [2] attempting to get -20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array2[1:][2, -20]
    assert str(excinfo.value) == "in ListArray64 with identity [3] attempting to get -20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array2[1, 0, 20]
    assert str(excinfo.value) == "in ListArray64 with identity [1, 0] attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array2[1:][2, 0, 20]
    assert str(excinfo.value) == "in ListArray64 with identity [3, 0] attempting to get 20, index out of range"

    with pytest.raises(ValueError) as excinfo:
        array2[:, 1:][3, 0, 20]
    assert str(excinfo.value) == "in ListArray64 with identity [3, 1] attempting to get 20, index out of range"

def test_array():
    starts  = numpy.array([0, 3, 3, 5, 6])
    stops   = numpy.array([3, 3, 5, 6, 10])
    content = numpy.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    array   = awkward1.layout.ListArray64(awkward1.layout.Index64(starts),
                                          awkward1.layout.Index64(stops),
                                          awkward1.layout.NumpyArray(content))

    @numba.njit
    def f1(q):
        return q[[2, 0, 20, 1],]

    with pytest.raises(ValueError) as excinfo:
        f1(array)
    assert str(excinfo.value) == "in ak::RegularArray, indexing error"



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import json

import numba

import awkward1.layout
from .._numba import util

class LiteralTypeType(numba.types.Type):
    def __init__(self, literal_type):
        super(LiteralTypeType, self).__init__(name="ak::LiteralTypeType({0})".format(repr(literal_type)))
        self.literal_type = literal_type

@numba.extending.register_model(LiteralTypeType)
class LiteralTypeModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        super(LiteralTypeModel, self).__init__(dmm, fe_type, [])

@numba.extending.unbox(LiteralTypeType)
def unbox(tpe, obj, c):
    proxyout = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder)
    is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)

@numba.extending.box(LiteralTypeType)
def box(tpe, val, c):
    return c.pyapi.unserialize(c.pyapi.serialize_object(tpe.literal_type))

def typeof_literaltype(literal_type):
    return LiteralTypeType(literal_type)

@numba.extending.typeof_impl.register(awkward1.layout.UnknownType)
def typeof_UnknownType(val, c):
    return UnknownTypeType(val.parameters)

@numba.extending.typeof_impl.register(awkward1.layout.PrimitiveType)
def typeof_PrimitiveType(val, c):
    return PrimitiveTypeType(val.dtype, val.parameters)

@numba.extending.typeof_impl.register(awkward1.layout.RegularType)
def typeof_RegularType(val, c):
    return RegularTypeType(numba.typeof(val.type), val.parameters)

@numba.extending.typeof_impl.register(awkward1.layout.ListType)
def typeof_ListType(val, c):
    return ListTypeType(numba.typeof(val.type), val.parameters)

@numba.extending.typeof_impl.register(awkward1.layout.OptionType)
def typeof_OptionType(val, c):
    return OptionTypeType(numba.typeof(val.type), val.parameters)

@numba.extending.typeof_impl.register(awkward1.layout.UnionType)
def typeof_UnionType(val, c):
    return UnionTypeType([numba.typeof(x) for x in val.types], val.parameters)

@numba.extending.typeof_impl.register(awkward1.layout.RecordType)
def typeof_RecordType(val, c):
    return RecordTypeType([numba.typeof(x) for x in val.types], None if val.istuple else tuple(val.keys()), val.parameters)

class TypeType(numba.types.Type):
    pass

class UnknownTypeType(TypeType):
    def __init__(self, parameters):
        super(UnknownTypeType, self).__init__(name="ak::UnknownTypeType(parameters={0})".format(json.dumps(parameters)))
        self.parameters = parameters

class PrimitiveTypeType(TypeType):
    def __init__(self, dtype, parameters):
        super(PrimitiveTypeType, self).__init__(name="ak::PrimitiveTypeType({0}, parameters={1})".format(dtype, json.dumps(parameters)))
        self.dtype = dtype
        self.parameters = parameters

class RegularTypeType(TypeType):
    def __init__(self, typetpe, parameters):
        super(RegularTypeType, self).__init__(name="ak::RegularTypeType({0}, parameters={1})".format(typetpe.name, json.dumps(parameters)))
        self.typetpe = typetpe
        self.parameters = parameters

class ListTypeType(TypeType):
    def __init__(self, typetpe, parameters):
        super(ListTypeType, self).__init__(name="ak::ListTypeType({0}, parameters={1})".format(typetpe.name, json.dumps(parameters)))
        self.typetpe = typetpe
        self.parameters = parameters

class OptionTypeType(TypeType):
    def __init__(self, typetpe, parameters):
        super(OptionTypeType, self).__init__(name="ak::OptionTypeType({0}, parameters={1})".format(typetpe.name, json.dumps(parameters)))
        self.typetpe = typetpe
        self.parameters = parameters

class UnionTypeType(TypeType):
    def __init__(self, typetpes, parameters):
        super(UnionTypeType, self).__init__(name="ak::UnionTypeType([{0}], parameters={1})".format(", ".join(x.name for x in typetpes), json.dumps(parameters)))
        self.typetpes = typetpes
        self.parameters = parameters

class RecordTypeType(TypeType):
    def __init__(self, typetpes, keys, parameters):
        super(RecordTypeType, self).__init__(name="ak::RecordTypeType([{0}], {1}, parameters={2})".format(", ".join(x.name for x in typetpes), repr(keys), json.dumps(parameters)))
        self.typetpes = typetpes
        self.keys = keys
        self.parameters = parameters

@numba.extending.register_model(UnknownTypeType)
class UnknownTypeModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        super(UnknownTypeModel, self).__init__(dmm, fe_type, [])

@numba.extending.register_model(PrimitiveTypeType)
class PrimitiveTypeModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        super(PrimitiveTypeModel, self).__init__(dmm, fe_type, [])

@numba.extending.register_model(RegularTypeType)
class RegularTypeModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [("type", fe_type.typetpe),
                   ("size", numba.int64)]
        super(RegularTypeModel, self).__init__(dmm, fe_type, members)

numba.extending.make_attribute_wrapper(RegularTypeType, "type", "type")

@numba.extending.register_model(ListTypeType)
class ListTypeModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [("type", fe_type.typetpe)]
        super(ListTypeModel, self).__init__(dmm, fe_type, members)

numba.extending.make_attribute_wrapper(ListTypeType, "type", "type")

@numba.extending.register_model(OptionTypeType)
class OptionTypeModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [("type", fe_type.typetpe)]
        super(OptionTypeModel, self).__init__(dmm, fe_type, members)

numba.extending.make_attribute_wrapper(OptionTypeType, "type", "type")

def field(i):
    return "f" + str(i)

@numba.extending.register_model(UnionTypeType)
class UnionTypeModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = []
        for i, tpe in enumerate(fe_type.typetpes):
            members.append((field(i), tpe))
        super(UnionTypeModel, self).__init__(dmm, fe_type, members)

@numba.extending.register_model(RecordTypeType)
class RecordTypeModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = []
        for i, tpe in enumerate(fe_type.typetpes):
            members.append((field(i), tpe))
        super(RecordTypeModel, self).__init__(dmm, fe_type, members)

@numba.extending.unbox(UnknownTypeType)
def unbox_UnknownType(tpe, obj, c):
    proxyout = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder)
    is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)

@numba.extending.unbox(PrimitiveTypeType)
def unbox_PrimitiveType(tpe, obj, c):
    proxyout = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder)
    is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)

@numba.extending.unbox(RegularTypeType)
def unbox_RegularType(tpe, obj, c):
    proxyout = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder)
    type_obj = c.pyapi.object_getattr_string(obj, "type")
    size_obj = c.pyapi.object_getattr_string(obj, "size")
    proxyout.type = c.pyapi.to_native_value(tpe.typetpe, type_obj).value
    proxyout.size = c.pyapi.to_native_value(numba.int64, size_obj).value
    c.pyapi.decref(type_obj)
    c.pyapi.decref(size_obj)
    is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)

@numba.extending.unbox(ListTypeType)
def unbox_ListType(tpe, obj, c):
    proxyout = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder)
    type_obj = c.pyapi.object_getattr_string(obj, "type")
    proxyout.type = c.pyapi.to_native_value(tpe.typetpe, type_obj).value
    c.pyapi.decref(type_obj)
    is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)

@numba.extending.unbox(OptionTypeType)
def unbox_OptionType(tpe, obj, c):
    proxyout = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder)
    type_obj = c.pyapi.object_getattr_string(obj, "type")
    proxyout.type = c.pyapi.to_native_value(tpe.typetpe, type_obj).value
    c.pyapi.decref(type_obj)
    is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)

@numba.extending.unbox(UnionTypeType)
def unbox_UnionType(tpe, obj, c):
    proxyout = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder)
    type_obj = c.pyapi.object_getattr_string(obj, "type")
    for i, t in enumerate(tpe.typetpes):
        i_obj = c.pyapi.long_from_longlong(c.context.get_constant(numba.int64, i))
        t_obj = c.pyapi.call_function_objargs(type_obj, (i_obj,))
        setattr(proxyout, field(i), c.pyapi.to_native_value(t, t_obj).value)
        c.pyapi.decref(i_obj)
        c.pyapi.decref(t_obj)
    c.pyapi.decref(type_obj)
    is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)

@numba.extending.unbox(RecordTypeType)
def unbox_RecordType(tpe, obj, c):
    proxyout = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder)
    field_obj = c.pyapi.object_getattr_string(obj, "field")
    for i, t in enumerate(tpe.typetpes):
        i_obj = c.pyapi.long_from_longlong(c.context.get_constant(numba.int64, i))
        t_obj = c.pyapi.call_function_objargs(field_obj, (i_obj,))
        setattr(proxyout, field(i), c.pyapi.to_native_value(t, t_obj).value)
        c.pyapi.decref(i_obj)
        c.pyapi.decref(t_obj)
    c.pyapi.decref(field_obj)
    is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)

@numba.extending.box(UnknownTypeType)
def box_UnknownType(tpe, val, c):
    return c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.UnknownType(tpe.parameters)))

@numba.extending.box(PrimitiveTypeType)
def box_PrimitiveType(tpe, val, c):
    return c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.PrimitiveType(tpe.dtype, tpe.parameters)))

def box_parameters(parameters, c):
    jsonloads_obj = c.pyapi.unserialize(c.pyapi.serialize_object(json.loads))
    paramstr_obj = c.pyapi.unserialize(c.pyapi.serialize_object(json.dumps(parameters)))
    param_obj = c.pyapi.call_function_objargs(jsonloads_obj, (paramstr_obj,))
    c.pyapi.decref(jsonloads_obj)
    c.pyapi.decref(paramstr_obj)
    return param_obj

@numba.extending.box(RegularTypeType)
def box_RegularType(tpe, val, c):
    proxyin = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder, value=val)
    class_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.RegularType))
    type_obj = c.pyapi.from_native_value(tpe.typetpe, proxyin.type, c.env_manager)
    size_obj = c.pyapi.long_from_longlong(proxyin.size)
    parameters_obj = box_parameters(tpe.parameters, c)
    out = c.pyapi.call_function_objargs(class_obj, (type_obj, size_obj, parameters_obj))
    c.pyapi.decref(class_obj)
    c.pyapi.decref(type_obj)
    c.pyapi.decref(size_obj)
    c.pyapi.decref(parameters_obj)
    return out

@numba.extending.box(ListTypeType)
def box_ListType(tpe, val, c):
    proxyin = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder, value=val)
    class_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.ListType))
    type_obj = c.pyapi.from_native_value(tpe.typetpe, proxyin.type, c.env_manager)
    parameters_obj = box_parameters(tpe.parameters, c)
    out = c.pyapi.call_function_objargs(class_obj, (type_obj, parameters_obj))
    c.pyapi.decref(class_obj)
    c.pyapi.decref(type_obj)
    c.pyapi.decref(parameters_obj)
    return out

@numba.extending.box(OptionTypeType)
def box_OptionType(tpe, val, c):
    proxyin = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder, value=val)
    class_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.OptionType))
    type_obj = c.pyapi.from_native_value(tpe.typetpe, proxyin.type, c.env_manager)
    parameters_obj = box_parameters(tpe.parameters, c)
    out = c.pyapi.call_function_objargs(class_obj, (type_obj, parameters_obj))
    c.pyapi.decref(class_obj)
    c.pyapi.decref(type_obj)
    c.pyapi.decref(parameters_obj)
    return out

@numba.extending.box(UnionTypeType)
def box_UnionType(tpe, val, c):
    proxyin = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder, value=val)
    class_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.UnionType))
    types_obj = c.pyapi.tuple_new(len(tpe.typetpes))
    for i, t in enumerate(tpe.typetpes):
        x_obj = c.pyapi.from_native_value(t, getattr(proxyin, field(i)), c.env_manager)
        c.pyapi.tuple_setitem(types_obj, i, x_obj)
    parameters_obj = box_parameters(tpe.parameters, c)
    out = c.pyapi.call_function_objargs(class_obj, (types_obj, parameters_obj))
    c.pyapi.decref(class_obj)
    c.pyapi.decref(types_obj)
    c.pyapi.decref(parameters_obj)
    return out

@numba.extending.box(RecordTypeType)
def box_RecordType(tpe, val, c):
    proxyin = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder, value=val)
    class_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.RecordType))
    types_obj = c.pyapi.tuple_new(len(tpe.typetpes))
    for i, t in enumerate(tpe.typetpes):
        x_obj = c.pyapi.from_native_value(t, getattr(proxyin, field(i)), c.env_manager)
        c.pyapi.tuple_setitem(types_obj, i, x_obj)
    parameters_obj = box_parameters(tpe.parameters, c)

    keys_obj = c.pyapi.unserialize(c.pyapi.serialize_object(tpe.keys))
    out = c.pyapi.call_function_objargs(class_obj, (types_obj, keys_obj, parameters_obj))

    c.pyapi.decref(class_obj)
    c.pyapi.decref(types_obj)
    c.pyapi.decref(keys_obj)
    c.pyapi.decref(parameters_obj)
    return out



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import operator

import numpy
import numba
import numba.typing.arraydecl

import awkward1.layout
from ..._numba import cpu, util, content

@numba.extending.typeof_impl.register(awkward1.layout.RegularArray)
def typeof(val, c):
    return RegularArrayType(numba.typeof(val.content), numba.typeof(val.identities), util.dict2parameters(val.parameters))

class RegularArrayType(content.ContentType):
    def __init__(self, contenttpe, identitiestpe, parameters):
        assert isinstance(parameters, tuple)
        super(RegularArrayType, self).__init__(name="ak::RegularArrayType({0}, identities={1}, parameters={2})".format(contenttpe.name, identitiestpe.name, util.parameters2str(parameters)))
        self.contenttpe = contenttpe
        self.identitiestpe = identitiestpe
        self.parameters = parameters

    @property
    def ndim(self):
        return 1 + self.contenttpe.ndim

    def getitem_int(self):
        return self.contenttpe

    def getitem_range(self):
        return self

    def getitem_str(self, key):
        return RegularArrayType(self.contenttpe.getitem_str(key), self.identitiestpe, ())

    def getitem_tuple(self, wheretpe):
        nexttpe = RegularArrayType(self, numba.none, ())
        out = nexttpe.getitem_next(wheretpe, False)
        return out.getitem_int()

    def getitem_next(self, wheretpe, isadvanced):
        if len(wheretpe.types) == 0:
            return self
        headtpe = wheretpe.types[0]
        tailtpe = numba.types.Tuple(wheretpe.types[1:])

        if isinstance(headtpe, numba.types.Integer):
            return self.contenttpe.carry().getitem_next(tailtpe, isadvanced)

        elif isinstance(headtpe, numba.types.SliceType):
            contenttpe = self.contenttpe.carry().getitem_next(tailtpe, isadvanced)
            return RegularArrayType(contenttpe, self.identitiestpe, self.parameters)

        elif isinstance(headtpe, numba.types.StringLiteral):
            return self.getitem_str(headtpe.literal_value).getitem_next(tailtpe, isadvanced)

        elif isinstance(headtpe, numba.types.EllipsisType):
            raise NotImplementedError("ellipsis")

        elif isinstance(headtpe, type(numba.typeof(numpy.newaxis))):
            raise NotImplementedError("newaxis")

        elif isinstance(headtpe, numba.types.Array):
            if headtpe.ndim != 1:
                raise NotImplementedError("array.ndim != 1")
            contenttpe = self.contenttpe.carry().getitem_next(tailtpe, True)
            if not isadvanced:
                return RegularArrayType(contenttpe, self.identitiestpe, self.parameters)
            else:
                return contenttpe

        else:
            raise AssertionError(headtpe)

    def carry(self):
        return RegularArrayType(self.contenttpe.carry(), self.identitiestpe, self.parameters)

    @property
    def lower_len(self):
        return lower_len

    @property
    def lower_getitem_nothing(self):
        return content.lower_getitem_nothing

    @property
    def lower_getitem_int(self):
        return lower_getitem_int

    @property
    def lower_getitem_range(self):
        return lower_getitem_range

    @property
    def lower_getitem_str(self):
        return lower_getitem_str

    @property
    def lower_getitem_next(self):
        return lower_getitem_next

    @property
    def lower_carry(self):
        return lower_carry

@numba.extending.register_model(RegularArrayType)
class RegularArrayModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [("content", fe_type.contenttpe),
                   ("size", numba.int64)]
        if fe_type.identitiestpe != numba.none:
            members.append(("identities", fe_type.identitiestpe))
        super(RegularArrayModel, self).__init__(dmm, fe_type, members)

@numba.extending.unbox(RegularArrayType)
def unbox(tpe, obj, c):
    content_obj = c.pyapi.object_getattr_string(obj, "content")
    size_obj = c.pyapi.object_getattr_string(obj, "size")
    proxyout = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder)
    proxyout.content = c.pyapi.to_native_value(tpe.contenttpe, content_obj).value
    proxyout.size = c.pyapi.to_native_value(numba.int64, size_obj).value
    c.pyapi.decref(content_obj)
    c.pyapi.decref(size_obj)
    if tpe.identitiestpe != numba.none:
        id_obj = c.pyapi.object_getattr_string(obj, "identities")
        proxyout.identities = c.pyapi.to_native_value(tpe.identitiestpe, id_obj).value
        c.pyapi.decref(id_obj)
    is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)

@numba.extending.box(RegularArrayType)
def box(tpe, val, c):
    RegularArray_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.RegularArray))
    proxyin = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder, value=val)
    content_obj = c.pyapi.from_native_value(tpe.contenttpe, proxyin.content, c.env_manager)
    size_obj = c.pyapi.long_from_longlong(proxyin.size)
    args = [content_obj, size_obj]
    if tpe.identitiestpe != numba.none:
        args.append(c.pyapi.from_native_value(tpe.identitiestpe, proxyin.identities, c.env_manager))
    else:
        args.append(c.pyapi.make_none())
    args.append(util.parameters2dict_impl(c, tpe.parameters))
    out = c.pyapi.call_function_objargs(RegularArray_obj, args)
    for x in args:
        c.pyapi.decref(x)
    c.pyapi.decref(RegularArray_obj)
    return out

@numba.extending.lower_builtin(len, RegularArrayType)
def lower_len(context, builder, sig, args):
    rettpe, (tpe,) = sig.return_type, sig.args
    val, = args
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    innerlen = tpe.contenttpe.lower_len(context, builder, rettpe(tpe.contenttpe), (proxyin.content,))
    size = util.cast(context, builder, numba.int64, numba.intp, proxyin.size)
    return builder.sdiv(innerlen, size)

@numba.extending.lower_builtin(operator.getitem, RegularArrayType, numba.types.Integer)
def lower_getitem_int(context, builder, sig, args):
    rettpe, (tpe, wheretpe) = sig.return_type, sig.args
    val, whereval = args
    whereval, wheretpe = util.cast(context, builder, wheretpe, numba.int64, whereval), numba.int64
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)

    start = builder.mul(whereval, proxyin.size)
    stop = builder.mul(builder.add(whereval, context.get_constant(numba.int64, 1)), proxyin.size)
    proxyslice = numba.cgutils.create_struct_proxy(numba.types.slice2_type)(context, builder)
    proxyslice.start = util.cast(context, builder, numba.int64, numba.intp, start)
    proxyslice.stop = util.cast(context, builder, numba.int64, numba.intp, stop)
    proxyslice.step = context.get_constant(numba.intp, 1)

    outtpe = tpe.contenttpe.getitem_range()
    return tpe.contenttpe.lower_getitem_range(context, builder, outtpe(tpe.contenttpe, numba.types.slice2_type), (proxyin.content, proxyslice._getvalue()))

@numba.extending.lower_builtin(operator.getitem, RegularArrayType, numba.types.slice2_type)
def lower_getitem_range(context, builder, sig, args):
    import awkward1._numba.identities

    rettpe, (tpe, wheretpe) = sig.return_type, sig.args
    val, whereval = args

    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    size = util.cast(context, builder, numba.int64, numba.intp, proxyin.size)

    proxyslicein = context.make_helper(builder, wheretpe, value=whereval)
    numba.targets.slicing.guard_invalid_slice(context, builder, wheretpe, proxyslicein)
    numba.targets.slicing.fix_slice(builder, proxyslicein, util.arraylen(context, builder, tpe, val))

    proxysliceout = numba.cgutils.create_struct_proxy(numba.types.slice2_type)(context, builder)
    proxysliceout.start = builder.mul(proxyslicein.start, size)
    proxysliceout.stop = builder.mul(proxyslicein.stop, size)
    proxysliceout.step = context.get_constant(numba.intp, 1)

    proxyout = numba.cgutils.create_struct_proxy(tpe)(context, builder)
    proxyout.content = tpe.contenttpe.lower_getitem_range(context, builder, rettpe.contenttpe(tpe.contenttpe, numba.types.slice2_type), (proxyin.content, proxysliceout._getvalue()))
    proxyout.size = proxyin.size
    if tpe.identitiestpe != numba.none:
        proxyout.identities = awkward1._numba.identities.lower_getitem_any(context, builder, tpe.identitiestpe, wheretpe, proxyin.identities, whereval)

    out = proxyout._getvalue()
    if context.enable_nrt:
        context.nrt.incref(builder, rettpe, out)
    return out

@numba.extending.lower_builtin(operator.getitem, RegularArrayType, numba.types.StringLiteral)
def lower_getitem_str(context, builder, sig, args):
    rettpe, (tpe, wheretpe) = sig.return_type, sig.args
    val, whereval = args

    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    proxyout = numba.cgutils.create_struct_proxy(rettpe)(context, builder)
    proxyout.size = proxyin.size
    proxyout.content = tpe.contenttpe.lower_getitem_str(context, builder, rettpe.contenttpe(tpe.contenttpe, wheretpe), (proxyin.content, whereval))
    if tpe.identitiestpe != numba.none:
        proxyout.identities = proxyin.identities

    out = proxyout._getvalue()
    if context.enable_nrt:
        context.nrt.incref(builder, rettpe, out)
    return out

@numba.extending.lower_builtin(operator.getitem, RegularArrayType, numba.types.BaseTuple)
def lower_getitem_tuple(context, builder, sig, args):
    return content.lower_getitem_tuple(context, builder, sig, args)

@numba.extending.lower_builtin(operator.getitem, RegularArrayType, numba.types.Array)
@numba.extending.lower_builtin(operator.getitem, RegularArrayType, numba.types.List)
@numba.extending.lower_builtin(operator.getitem, RegularArrayType, numba.types.ArrayCompatible)
@numba.extending.lower_builtin(operator.getitem, RegularArrayType, numba.types.EllipsisType)
@numba.extending.lower_builtin(operator.getitem, RegularArrayType, type(numba.typeof(numpy.newaxis)))
def lower_getitem_other(context, builder, sig, args):
    return content.lower_getitem_other(context, builder, sig, args)

def lower_getitem_next(context, builder, arraytpe, wheretpe, arrayval, whereval, advanced):
    if len(wheretpe.types) == 0:
        return arrayval

    headtpe = wheretpe.types[0]
    tailtpe = numba.types.Tuple(wheretpe.types[1:])
    headval = numba.cgutils.unpack_tuple(builder, whereval)[0]
    tailval = context.make_tuple(builder, tailtpe, numba.cgutils.unpack_tuple(builder, whereval)[1:])

    proxyin = numba.cgutils.create_struct_proxy(arraytpe)(context, builder, value=arrayval)
    leng = util.arraylen(context, builder, arraytpe, arrayval, totpe=numba.int64)

    if isinstance(headtpe, numba.types.Integer):
        assert advanced is None
        nextcarry = util.newindex64(context, builder, numba.int64, leng)
        util.call(context, builder, cpu.kernels.awkward_regulararray_getitem_next_at_64,
            (util.arrayptr(context, builder, util.index64tpe, nextcarry),
             util.cast(context, builder, headtpe, numba.int64, headval),
             leng,
             proxyin.size),
            "in {0}, indexing error".format(arraytpe.shortname))
        nextcontenttpe = arraytpe.contenttpe.carry()
        nextcontentval = arraytpe.contenttpe.lower_carry(context, builder, arraytpe.contenttpe, util.index64tpe, proxyin.content, nextcarry)
        return nextcontenttpe.lower_getitem_next(context, builder, nextcontenttpe, tailtpe, nextcontentval, tailval, advanced)

    elif isinstance(headtpe, numba.types.SliceType):
        proxyslicein = context.make_helper(builder, headtpe, value=headval)
        numba.targets.slicing.guard_invalid_slice(context, builder, headtpe, proxyslicein)
        numba.targets.slicing.fix_slice(builder, proxyslicein, util.cast(context, builder, numba.int64, numba.intp, proxyin.size))
        nextsize = util.cast(context, builder, numba.intp, numba.int64, numba.targets.slicing.get_slice_length(builder, proxyslicein))

        nextcarry = util.newindex64(context, builder, numba.int64, builder.mul(leng, nextsize))
        util.call(context, builder, cpu.kernels.awkward_regulararray_getitem_next_range_64,
            (util.arrayptr(context, builder, util.index64tpe, nextcarry),
             util.cast(context, builder, numba.intp, numba.int64, proxyslicein.start),
             util.cast(context, builder, numba.intp, numba.int64, proxyslicein.step),
             leng,
             proxyin.size,
             nextsize),
            "in {0}, indexing error".format(arraytpe.shortname))

        nextcontenttpe = arraytpe.contenttpe.carry()
        nextcontentval = arraytpe.contenttpe.lower_carry(context, builder, arraytpe.contenttpe, util.index64tpe, proxyin.content, nextcarry)

        if advanced is None:
            outcontenttpe = nextcontenttpe.getitem_next(tailtpe, False)
            outcontentval = nextcontenttpe.lower_getitem_next(context, builder, nextcontenttpe, tailtpe, nextcontentval, tailval, advanced)

        else:
            nextadvanced = util.newindex64(context, builder, numba.int64, builder.mul(leng, nextsize))
            util.call(context, builder, cpu.kernels.awkward_regulararray_getitem_next_range_spreadadvanced_64,
                (util.arrayptr(context, builder, util.index64tpe, nextadvanced),
                 util.arrayptr(context, builder, util.index64tpe, advanced),
                 leng,
                 nextsize),
                "in {0}, indexing error".format(arraytpe.shortname))

            outcontenttpe = nextcontenttpe.getitem_next(tailtpe, True)
            outcontentval = nextcontenttpe.lower_getitem_next(context, builder, nextcontenttpe, tailtpe, nextcontentval, tailval, nextadvanced)

        outtpe = RegularArrayType(outcontenttpe, arraytpe.identitiestpe, arraytpe.parameters)
        proxyout = numba.cgutils.create_struct_proxy(outtpe)(context, builder)
        proxyout.content = outcontentval
        proxyout.size = nextsize
        if arraytpe.identitiestpe != numba.none:
            proxyout.identities = proxyin.identities
        return proxyout._getvalue()

    elif isinstance(headtpe, numba.types.StringLiteral):
        nexttpe = arraytpe.getitem_str(headtpe.literal_value)
        nextval = lower_getitem_str(context, builder, nexttpe(arraytpe, headtpe), (arrayval, headval))
        return lower_getitem_next(context, builder, nexttpe, tailtpe, nextval, tailval, advanced)

    elif isinstance(headtpe, numba.types.EllipsisType):
        raise NotImplementedError("RegularArray.getitem_next(ellipsis)")

    elif isinstance(headtpe, type(numba.typeof(numpy.newaxis))):
        raise NotImplementedError("RegularArray.getitem_next(newaxis)")

    elif isinstance(headtpe, numba.types.Array):
        if headtpe.ndim != 1:
            raise NotImplementedError("array.ndim != 1")

        flathead = numba.targets.arrayobj.array_ravel(context, builder, util.index64tpe(headtpe), (headval,))
        lenflathead = util.arraylen(context, builder, util.index64tpe, flathead, totpe=numba.int64)
        regular_flathead = util.newindex64(context, builder, numba.int64, lenflathead)

        util.call(context, builder, cpu.kernels.awkward_regulararray_getitem_next_array_regularize_64,
            (util.arrayptr(context, builder, util.index64tpe, regular_flathead),
             util.arrayptr(context, builder, util.index64tpe, flathead),
             lenflathead,
             proxyin.size),
            "in {0}, indexing error".format(arraytpe.shortname))

        if advanced is None:
            lencarry = builder.mul(leng, lenflathead)
            nextcarry = util.newindex64(context, builder, numba.int64, lencarry)
            nextadvanced = util.newindex64(context, builder, numba.int64, lencarry)

            util.call(context, builder, cpu.kernels.awkward_regulararray_getitem_next_array_64,
                (util.arrayptr(context, builder, util.index64tpe, nextcarry),
                 util.arrayptr(context, builder, util.index64tpe, nextadvanced),
                 util.arrayptr(context, builder, util.index64tpe, regular_flathead),
                 leng,
                 lenflathead,
                 proxyin.size),
                "in {0}, indexing error".format(arraytpe.shortname))

            nexttpe = arraytpe.contenttpe.carry()
            nextval = arraytpe.contenttpe.lower_carry(context, builder, arraytpe.contenttpe, util.index64tpe, proxyin.content, nextcarry)

            contenttpe = nexttpe.getitem_next(tailtpe, True)
            contentval = nexttpe.lower_getitem_next(context, builder, nexttpe, tailtpe, nextval, tailval, nextadvanced)

            outtpe = RegularArrayType(contenttpe, arraytpe.identitiestpe, arraytpe.parameters)
            proxyout = numba.cgutils.create_struct_proxy(outtpe)(context, builder)
            proxyout.content = contentval
            proxyout.size = lenflathead
            if outtpe.identitiestpe != numba.none:
                proxyout.identities = awkward1._numba.identities.lower_getitem_any(context, builder, outtpe.identitiestpe, util.index64tpe, proxyin.identities, flathead)
            return proxyout._getvalue()

        else:
            nextcarry = util.newindex64(context, builder, numba.int64, leng)
            nextadvanced = util.newindex64(context, builder, numba.int64, leng)

            util.call(context, builder, cpu.kernels.awkward_regulararray_getitem_next_array_advanced_64,
                (util.arrayptr(context, builder, util.index64tpe, nextcarry),
                 util.arrayptr(context, builder, util.index64tpe, nextadvanced),
                 util.arrayptr(context, builder, util.index64tpe, advanced),
                 util.arrayptr(context, builder, util.index64tpe, regular_flathead),
                 leng,
                 lenflathead,
                 proxyin.size),
                "in {0}, indexing error".format(arraytpe.shortname))

            nexttpe = arraytpe.contenttpe.carry()
            nextval = arraytpe.contenttpe.lower_carry(context, builder, arraytpe.contenttpe, util.index64tpe, proxyin.content, nextcarry)

            outtpe = nexttpe.getitem_next(tailtpe, True)
            return nexttpe.lower_getitem_next(context, builder, nexttpe, tailtpe, nextval, tailval, nextadvanced)

    else:
        raise AssertionError(headtpe)

def lower_carry(context, builder, arraytpe, carrytpe, arrayval, carryval):
    import awkward1._numba.identities

    proxyin = numba.cgutils.create_struct_proxy(arraytpe)(context, builder, value=arrayval)

    lencarry = util.arraylen(context, builder, carrytpe, carryval, totpe=numba.int64)
    nextcarry = util.newindex64(context, builder, numba.int64, builder.mul(lencarry, proxyin.size))
    util.call(context, builder, cpu.kernels.awkward_regulararray_getitem_carry_64,
        (util.arrayptr(context, builder, util.index64tpe, nextcarry),
         util.arrayptr(context, builder, carrytpe, carryval),
         lencarry,
         proxyin.size),
        "in {0}, indexing error".format(arraytpe.shortname))
    nextcontent = arraytpe.contenttpe.lower_carry(context, builder, arraytpe.contenttpe, util.index64tpe, proxyin.content, nextcarry)

    rettpe = arraytpe.carry()
    proxyout = numba.cgutils.create_struct_proxy(rettpe)(context, builder)
    proxyout.content = nextcontent
    proxyout.size = proxyin.size
    if rettpe.identitiestpe != numba.none:
        proxyout.identities = awkward1._numba.identities.lower_getitem_any(context, builder, rettpe.identitiestpe, carrytpe, proxyin.identities, carryval)

    return proxyout._getvalue()

@numba.typing.templates.infer_getattr
class type_methods(numba.typing.templates.AttributeTemplate):
    key = RegularArrayType

    def generic_resolve(self, tpe, attr):
        if attr == "content":
            return tpe.contenttpe

        elif attr == "size":
            return numba.int64

        elif attr == "identities":
            if tpe.identitiestpe == numba.none:
                return numba.optional(identity.IdentitiesType(numba.int32[:, :]))
            else:
                return tpe.identitiestpe

@numba.extending.lower_getattr(RegularArrayType, "content")
def lower_content(context, builder, tpe, val):
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    if context.enable_nrt:
        context.nrt.incref(builder, tpe.contenttpe, proxyin.content)
    return proxyin.content

@numba.extending.lower_getattr(RegularArrayType, "size")
def lower_size(context, builder, tpe, val):
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    return proxyin.size

@numba.extending.lower_getattr(RegularArrayType, "identities")
def lower_identities(context, builder, tpe, val):
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    if tpe.identitiestpe == numba.none:
        return context.make_optional_none(builder, identity.IdentitiesType(numba.int32[:, :]))
    else:
        if context.enable_nrt:
            context.nrt.incref(builder, tpe.identitiestpe, proxyin.identities)
        return proxyin.identities



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import os
import json

import pytest
import numpy
numba = pytest.importorskip("numba")

import awkward1

py27 = (sys.version_info[0] < 3)

def test_index():
    array_i1 = numpy.array([numpy.iinfo("i1").min, -1, 0, 1, numpy.iinfo("i1").max], dtype="i1")
    array_u1 = numpy.array([numpy.iinfo("u1").min, -1, 0, 1, numpy.iinfo("u1").max], dtype="u1")
    array_li2 = numpy.array([numpy.iinfo("<i2").min, -1, 0, 1, numpy.iinfo("<i2").max], dtype="<i2")
    array_lu2 = numpy.array([numpy.iinfo("<u2").min, -1, 0, 1, numpy.iinfo("<u2").max], dtype="<u2")
    array_li4 = numpy.array([numpy.iinfo("<i4").min, -1, 0, 1, numpy.iinfo("<i4").max], dtype="<i4")
    array_lu4 = numpy.array([numpy.iinfo("<u4").min, -1, 0, 1, numpy.iinfo("<u4").max], dtype="<u4")
    array_li8 = numpy.array([numpy.iinfo("<i8").min, -1, 0, 1, numpy.iinfo("<i8").max], dtype="<i8")
    array_lu8 = numpy.array([numpy.iinfo("<u8").min, -1, 0, 1, numpy.iinfo("<u8").max], dtype="<u8")
    array_bi2 = numpy.array([numpy.iinfo(">i2").min, -1, 0, 1, numpy.iinfo(">i2").max], dtype=">i2")
    array_bu2 = numpy.array([numpy.iinfo(">u2").min, -1, 0, 1, numpy.iinfo(">u2").max], dtype=">u2")
    array_bi4 = numpy.array([numpy.iinfo(">i4").min, -1, 0, 1, numpy.iinfo(">i4").max], dtype=">i4")
    array_bu4 = numpy.array([numpy.iinfo(">u4").min, -1, 0, 1, numpy.iinfo(">u4").max], dtype=">u4")
    array_bi8 = numpy.array([numpy.iinfo(">i8").min, -1, 0, 1, numpy.iinfo(">i8").max], dtype=">i8")
    array_bu8 = numpy.array([numpy.iinfo(">u8").min, -1, 0, 1, numpy.iinfo(">u8").max], dtype=">u8")

    index_i1 = awkward1.layout.Index8(array_i1)
    index_u1 = awkward1.layout.IndexU8(array_u1)
    index_li2 = awkward1.layout.Index32(array_li2)
    index_lu2 = awkward1.layout.Index32(array_lu2)
    index_li4 = awkward1.layout.Index32(array_li4)
    index_lu4 = awkward1.layout.IndexU32(array_lu4)
    index_li8 = awkward1.layout.Index64(array_li8)
    index_lu8 = awkward1.layout.Index64(array_lu8)
    index_bi2 = awkward1.layout.Index32(array_bi2)
    index_bu2 = awkward1.layout.Index32(array_bu2)
    index_bi4 = awkward1.layout.Index32(array_bi4)
    index_bu4 = awkward1.layout.IndexU32(array_bu4)
    index_bi8 = awkward1.layout.Index64(array_bi8)
    index_bu8 = awkward1.layout.Index64(array_bu8)

    assert index_i1[2] == 0
    assert index_u1[2] == 0
    assert index_li2[2] == 0
    assert index_lu2[2] == 0
    assert index_li4[2] == 0
    assert index_lu4[2] == 0
    assert index_li8[2] == 0
    assert index_lu8[2] == 0
    assert index_bi2[2] == 0
    assert index_bu2[2] == 0
    assert index_bi4[2] == 0
    assert index_bu4[2] == 0
    assert index_bi8[2] == 0
    assert index_bu8[2] == 0

    array_i1[2] = 10
    array_u1[2] = 10
    array_li2[2] = 10
    array_lu2[2] = 10
    array_li4[2] = 10
    array_lu4[2] = 10
    array_li8[2] = 10
    array_lu8[2] = 10
    array_bi2[2] = 10
    array_bu2[2] = 10
    array_bi4[2] = 10
    array_bu4[2] = 10
    array_bi8[2] = 10
    array_bu8[2] = 10

    assert index_i1[2] == 10
    assert index_u1[2] == 10
    assert index_li2[2] == 0
    assert index_lu2[2] == 0
    assert index_li4[2] == 10
    assert index_lu4[2] == 10
    assert index_li8[2] == 10
    assert index_lu8[2] == 0
    assert index_bi2[2] == 0
    assert index_bu2[2] == 0
    assert index_bi4[2] == 0
    assert index_bu4[2] == 0
    assert index_bi8[2] == 0
    assert index_bu8[2] == 0

content  = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
starts1  = awkward1.layout.IndexU32(numpy.array([0, 3, 3, 5, 6], numpy.uint32))
stops1   = awkward1.layout.IndexU32(numpy.array([3, 3, 5, 6, 9], numpy.uint32))
offsets1 = awkward1.layout.IndexU32(numpy.array([0, 3, 3, 5, 6, 9], numpy.uint32))
starts2  = awkward1.layout.IndexU32(numpy.array([0, 2, 3, 3], numpy.uint32))
stops2   = awkward1.layout.IndexU32(numpy.array([2, 3, 3, 5], numpy.uint32))
offsets2 = awkward1.layout.IndexU32(numpy.array([0, 2, 3, 3, 5], numpy.uint32))

def test_listarray_basic():
    array1 = awkward1.layout.ListArrayU32(starts1, stops1, content)
    array2 = awkward1.layout.ListArrayU32(starts2, stops2, array1)
    assert awkward1.tolist(array1) == [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert awkward1.tolist(array1[2]) == [4.4, 5.5]
    assert awkward1.tolist(array1[1:-1]) == [[], [4.4, 5.5], [6.6]]
    assert awkward1.tolist(array2) == [[[1.1, 2.2, 3.3], []], [[4.4, 5.5]], [], [[6.6], [7.7, 8.8, 9.9]]]
    assert awkward1.tolist(array2[1]) == [[4.4, 5.5]]
    assert awkward1.tolist(array2[1:-1]) == [[[4.4, 5.5]], []]

def test_listoffsetarray_basic():
    array1 = awkward1.layout.ListOffsetArrayU32(offsets1, content)
    array2 = awkward1.layout.ListOffsetArrayU32(offsets2, array1)
    assert awkward1.tolist(array1) == [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert awkward1.tolist(array1[2]) == [4.4, 5.5]
    assert awkward1.tolist(array1[1:-1]) == [[], [4.4, 5.5], [6.6]]
    assert awkward1.tolist(array2) == [[[1.1, 2.2, 3.3], []], [[4.4, 5.5]], [], [[6.6], [7.7, 8.8, 9.9]]]
    assert awkward1.tolist(array2[1]) == [[4.4, 5.5]]
    assert awkward1.tolist(array2[1:-1]) == [[[4.4, 5.5]], []]

def test_listarray_at():
    array1 = awkward1.layout.ListArrayU32(starts1, stops1, content)
    array2 = awkward1.layout.ListArrayU32(starts2, stops2, array1)
    assert awkward1.tolist(array1[2]) == [4.4, 5.5]
    assert awkward1.tolist(array1[2,]) == [4.4, 5.5]
    assert awkward1.tolist(array1[2, 1:]) == [5.5]
    assert awkward1.tolist(array1[2:, 0]) == [4.4, 6.6, 7.7]
    assert awkward1.tolist(array1[2:, -1]) == [5.5, 6.6, 9.9]

def test_listoffsetarray_at():
    array1 = awkward1.layout.ListOffsetArrayU32(offsets1, content)
    array2 = awkward1.layout.ListOffsetArrayU32(offsets2, array1)
    assert awkward1.tolist(array1[2,]) == [4.4, 5.5]
    assert awkward1.tolist(array1[2, 1:]) == [5.5]
    assert awkward1.tolist(array1[2:, 0]) == [4.4, 6.6, 7.7]
    assert awkward1.tolist(array1[2:, -1]) == [5.5, 6.6, 9.9]

def test_listarray_slice():
    array1 = awkward1.layout.ListArrayU32(starts1, stops1, content)
    array2 = awkward1.layout.ListArrayU32(starts2, stops2, array1)
    assert awkward1.tolist(array1[1:-1]) == [[], [4.4, 5.5], [6.6]]
    assert awkward1.tolist(array1[1:-1,]) == [[], [4.4, 5.5], [6.6]]
    assert awkward1.tolist(array2[1:-1]) == [[[4.4, 5.5]], []]
    assert awkward1.tolist(array2[1:-1,]) == [[[4.4, 5.5]], []]

def test_listoffsetarray_slice():
    array1 = awkward1.layout.ListOffsetArrayU32(offsets1, content)
    array2 = awkward1.layout.ListOffsetArrayU32(offsets2, array1)
    assert awkward1.tolist(array1[1:-1]) == [[], [4.4, 5.5], [6.6]]
    assert awkward1.tolist(array1[1:-1,]) == [[], [4.4, 5.5], [6.6]]
    assert awkward1.tolist(array2[1:-1]) == [[[4.4, 5.5]], []]
    assert awkward1.tolist(array2[1:-1,]) == [[[4.4, 5.5]], []]

def test_listarray_slice_slice():
    array1 = awkward1.layout.ListArrayU32(starts1, stops1, content)
    array2 = awkward1.layout.ListArrayU32(starts2, stops2, array1)
    assert awkward1.tolist(array1[2:]) == [[4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert awkward1.tolist(array1[2:, 1:]) == [[5.5], [], [8.8, 9.9]]
    assert awkward1.tolist(array1[2:,:-1]) == [[4.4], [], [7.7, 8.8]]

def test_listoffsetarray_slice_slice():
    array1 = awkward1.layout.ListOffsetArrayU32(offsets1, content)
    array2 = awkward1.layout.ListOffsetArrayU32(offsets2, array1)
    assert awkward1.tolist(array1[2:]) == [[4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert awkward1.tolist(array1[2:, 1:]) == [[5.5], [], [8.8, 9.9]]
    assert awkward1.tolist(array1[2:,:-1]) == [[4.4], [], [7.7, 8.8]]

def test_listarray_ellipsis():
    array1 = awkward1.layout.ListArrayU32(starts1, stops1, content)
    array2 = awkward1.layout.ListArrayU32(starts2, stops2, array1)
    if not py27:
        assert awkward1.tolist(array1[Ellipsis, 1:]) == [[2.2, 3.3], [], [5.5], [], [8.8, 9.9]]
        assert awkward1.tolist(array2[Ellipsis, 1:]) == [[[2.2, 3.3], []], [[5.5]], [], [[], [8.8, 9.9]]]

def test_listoffsetarray_ellipsis():
    array1 = awkward1.layout.ListOffsetArrayU32(offsets1, content)
    array2 = awkward1.layout.ListOffsetArrayU32(offsets2, array1)
    if not py27:
        assert awkward1.tolist(array1[Ellipsis, 1:]) == [[2.2, 3.3], [], [5.5], [], [8.8, 9.9]]
        assert awkward1.tolist(array2[Ellipsis, 1:]) == [[[2.2, 3.3], []], [[5.5]], [], [[], [8.8, 9.9]]]

def test_listarray_array_slice():
    array1 = awkward1.layout.ListArrayU32(starts1, stops1, content)
    array2 = awkward1.layout.ListArrayU32(starts2, stops2, array1)
    assert awkward1.tolist(array2[[0, 0, 1, 1, 1, 0]]) == [[[1.1, 2.2, 3.3], []], [[1.1, 2.2, 3.3], []], [[4.4, 5.5]], [[4.4, 5.5]], [[4.4, 5.5]], [[1.1, 2.2, 3.3], []]]
    assert awkward1.tolist(array2[[0, 0, 1, 1, 1, 0], :]) == [[[1.1, 2.2, 3.3], []], [[1.1, 2.2, 3.3], []], [[4.4, 5.5]], [[4.4, 5.5]], [[4.4, 5.5]], [[1.1, 2.2, 3.3], []]]
    assert awkward1.tolist(array2[[0, 0, 1, 1, 1, 0], :, 1:]) == [[[2.2, 3.3], []], [[2.2, 3.3], []], [[5.5]], [[5.5]], [[5.5]], [[2.2, 3.3], []]]

def test_listoffsetarray_array_slice():
    array1 = awkward1.layout.ListOffsetArrayU32(offsets1, content)
    array2 = awkward1.layout.ListOffsetArrayU32(offsets2, array1)
    assert awkward1.tolist(array2[[0, 0, 1, 1, 1, 0]]) == [[[1.1, 2.2, 3.3], []], [[1.1, 2.2, 3.3], []], [[4.4, 5.5]], [[4.4, 5.5]], [[4.4, 5.5]], [[1.1, 2.2, 3.3], []]]
    assert awkward1.tolist(array2[[0, 0, 1, 1, 1, 0], :]) == [[[1.1, 2.2, 3.3], []], [[1.1, 2.2, 3.3], []], [[4.4, 5.5]], [[4.4, 5.5]], [[4.4, 5.5]], [[1.1, 2.2, 3.3], []]]
    assert awkward1.tolist(array2[[0, 0, 1, 1, 1, 0], :, 1:]) == [[[2.2, 3.3], []], [[2.2, 3.3], []], [[5.5]], [[5.5]], [[5.5]], [[2.2, 3.3], []]]

def test_listarray_array():
    array1 = awkward1.layout.ListArrayU32(starts1, stops1, content)
    array2 = awkward1.layout.ListArrayU32(starts2, stops2, array1)
    assert awkward1.tolist(array1[numpy.array([2, 0, 0, 1, -1])]) == [[4.4, 5.5], [1.1, 2.2, 3.3], [1.1, 2.2, 3.3], [], [7.7, 8.8, 9.9]]
    assert awkward1.tolist(array1[numpy.array([2, 0, 0, -1]), numpy.array([1, 1, 0, 0])]) == [5.5, 2.2, 1.1, 7.7]

    content_deep = awkward1.layout.NumpyArray(numpy.array([[0, 0], [1, 10], [2, 20], [3, 30], [4, 40], [5, 50], [6, 60], [7, 70], [8, 80]]))
    starts1_deep = awkward1.layout.IndexU32(numpy.array([0, 3, 6]))
    stops1_deep = awkward1.layout.IndexU32(numpy.array([3, 6, 9]))
    array1_deep = awkward1.layout.ListArrayU32(starts1_deep, stops1_deep, content_deep)

    assert awkward1.tolist(array1_deep) == [[[0, 0], [1, 10], [2, 20]], [[3, 30], [4, 40], [5, 50]], [[6, 60], [7, 70], [8, 80]]]
    s = (numpy.array([2, 0, 0, -1]), numpy.array([1, 1, 0, 0]), numpy.array([0, 1, 0, 1]))
    assert numpy.array([[[0, 0], [1, 10], [2, 20]], [[3, 30], [4, 40], [5, 50]], [[6, 60], [7, 70], [8, 80]]])[s].tolist() == awkward1.tolist(array1_deep[s])

    s = (numpy.array([2, 0, 0, -1]), numpy.array([1, 1, 0, 0]), slice(1, None))
    assert numpy.array([[[0, 0], [1, 10], [2, 20]], [[3, 30], [4, 40], [5, 50]], [[6, 60], [7, 70], [8, 80]]])[s].tolist() == awkward1.tolist(array1_deep[s])

def test_listoffsetarray_array():
    array1 = awkward1.layout.ListOffsetArrayU32(offsets1, content)
    array2 = awkward1.layout.ListOffsetArrayU32(offsets2, array1)
    assert awkward1.tolist(array1[numpy.array([2, 0, 0, 1, -1])]) == [[4.4, 5.5], [1.1, 2.2, 3.3], [1.1, 2.2, 3.3], [], [7.7, 8.8, 9.9]]
    assert awkward1.tolist(array1[numpy.array([2, 0, 0, -1]), numpy.array([1, 1, 0, 0])]) == [5.5, 2.2, 1.1, 7.7]

    content_deep = awkward1.layout.NumpyArray(numpy.array([[0, 0], [1, 10], [2, 20], [3, 30], [4, 40], [5, 50], [6, 60], [7, 70], [8, 80]]))
    starts1_deep = awkward1.layout.IndexU32(numpy.array([0, 3, 6]))
    stops1_deep = awkward1.layout.IndexU32(numpy.array([3, 6, 9]))
    array1_deep = awkward1.layout.ListArrayU32(starts1_deep, stops1_deep, content_deep)

    assert awkward1.tolist(array1_deep) == [[[0, 0], [1, 10], [2, 20]], [[3, 30], [4, 40], [5, 50]], [[6, 60], [7, 70], [8, 80]]]
    s = (numpy.array([2, 0, 0, -1]), numpy.array([1, 1, 0, 0]), numpy.array([0, 1, 0, 1]))
    assert numpy.array([[[0, 0], [1, 10], [2, 20]], [[3, 30], [4, 40], [5, 50]], [[6, 60], [7, 70], [8, 80]]])[s].tolist() == awkward1.tolist(array1_deep[s])

    s = (numpy.array([2, 0, 0, -1]), numpy.array([1, 1, 0, 0]), slice(1, None))
    assert numpy.array([[[0, 0], [1, 10], [2, 20]], [[3, 30], [4, 40], [5, 50]], [[6, 60], [7, 70], [8, 80]]])[s].tolist() == awkward1.tolist(array1_deep[s])

def test_getitem_array_U32():
    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
    listarray = awkward1.layout.ListArrayU32(awkward1.layout.IndexU32(numpy.array([0, 3, 3, 5, 6], numpy.uint32)), awkward1.layout.IndexU32(numpy.array([3, 3, 5, 6, 10], numpy.uint32)), content)

    @numba.njit
    def f1(q):
        return q[[2, 0, 0, 1]]

    assert awkward1.tolist(f1(listarray)) == [[3.3, 4.4], [0.0, 1.1, 2.2], [0.0, 1.1, 2.2], []]

    @numba.njit
    def f2(q):
        return q[[2, 0, 0, -1], 1]

    assert awkward1.tolist(f2(listarray)) == [4.4, 1.1, 1.1, 7.7]

    @numba.njit
    def f3(q):
        return q[[2, 0, 0, -1], [-1, -1, 0, 0]]

    assert awkward1.tolist(f3(listarray)) == [4.4, 2.2, 0.0, 6.6]

    listarray.setidentities()
    assert numpy.asarray(f1(listarray).identities).tolist() == [[2], [0], [0], [1]]
    assert numpy.asarray(f2(listarray).identities).tolist() == [[2, 1], [0, 1], [0, 1], [4, 1]]
    assert numpy.asarray(f3(listarray).identities).tolist() == [[2, 1], [0, 2], [0, 0], [4, 0]]

def test_deep_listarrayU32():
    content = awkward1.layout.NumpyArray(numpy.arange(2*3*5*7).reshape(-1, 7))
    offsetsA = numpy.arange(0, 2*3*5 + 5, 5)
    offsetsB = numpy.arange(0, 2*3 + 3, 3)
    startsA, stopsA = offsetsA[:-1], offsetsA[1:]
    startsB, stopsB = offsetsB[:-1], offsetsB[1:]

    listarrayAU32 = awkward1.layout.ListArrayU32(awkward1.layout.IndexU32(startsA), awkward1.layout.IndexU32(stopsA), content)
    modelA = numpy.arange(2*3*5*7).reshape(2*3, 5, 7)

    listarrayBU32 = awkward1.layout.ListArrayU32(awkward1.layout.IndexU32(startsB), awkward1.layout.IndexU32(stopsB), listarrayAU32)
    modelB = numpy.arange(2*3*5*7).reshape(2, 3, 5, 7)

    @numba.njit
    def f1(q):
        return q[1, 2, 4]

    assert awkward1.tolist(f1(listarrayBU32)) == awkward1.tolist(f1(modelB))

    @numba.njit
    def f2(q):
        return q[1, -1, 4, -2]

    assert f2(listarrayBU32) == f2(modelB)

    @numba.njit
    def f3(q):
        return q[1:, ::2, 1:-1, 5::-1]

    assert awkward1.tolist(f3(listarrayBU32)) == awkward1.tolist(f3(modelB))

    @numba.njit
    def f4(q):
        return q[numpy.array([0, 1, 1, 0]), ::2, 1:-1, 5::-1]

    assert awkward1.tolist(f4(listarrayBU32)) == awkward1.tolist(f4(modelB))

def test_deep_listoffsetarrayU32():
    content = awkward1.layout.NumpyArray(numpy.arange(2*3*5*7).reshape(-1, 7))
    offsetsA = numpy.arange(0, 2*3*5 + 5, 5)
    offsetsB = numpy.arange(0, 2*3 + 3, 3)

    listoffsetarrayAU32 = awkward1.layout.ListOffsetArrayU32(awkward1.layout.IndexU32(offsetsA), content)
    modelA = numpy.arange(2*3*5*7).reshape(2*3, 5, 7)

    listoffsetarrayBU32 = awkward1.layout.ListOffsetArrayU32(awkward1.layout.IndexU32(offsetsB), listoffsetarrayAU32)
    modelB = numpy.arange(2*3*5*7).reshape(2, 3, 5, 7)

    @numba.njit
    def f1(q):
        return q[1, 2, 4]

    assert awkward1.tolist(f1(listoffsetarrayBU32)) == awkward1.tolist(f1(modelB))

    @numba.njit
    def f2(q):
        return q[1, -1, 4, -2]

    assert f2(listoffsetarrayBU32) == f2(modelB)

    @numba.njit
    def f3(q):
        return q[1:, ::2, 1:-1, 5::-1]

    assert awkward1.tolist(f3(listoffsetarrayBU32)) == awkward1.tolist(f3(modelB))

    @numba.njit
    def f4(q):
        return q[numpy.array([0, 1, 1, 0]), ::2, 1:-1, 5::-1]

    assert awkward1.tolist(f4(listoffsetarrayBU32)) == awkward1.tolist(f4(modelB))



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import numpy

import awkward1.layout
import awkward1.operations.convert
import awkward1._util

def convert_to_array(layout, args, kwargs):
    out = awkward1.operations.convert.tonumpy(layout)
    if args == () and kwargs == {}:
        return out
    else:
        return numpy.array(out, *args, **kwargs)

implemented = {}

def array_function(func, types, args, kwargs):
    function = implemented.get(func)
    if function is None:
        return NotImplemented
    else:
        return function(*args, **kwargs)

def implements(numpy_function):
    def decorator(function):
        implemented[numpy_function] = function
        return function
    return decorator

def array_ufunc(ufunc, method, inputs, kwargs, classes, functions):
    import awkward1.highlevel

    if method != "__call__" or len(inputs) == 0 or "out" in kwargs:
        return NotImplemented

    def unwrap(x):
        if isinstance(x, (awkward1.highlevel.Array, awkward1.highlevel.Record)):
            return x.layout
        elif isinstance(x, awkward1.highlevel.FillableArray):
            return x.snapshot().layout
        elif isinstance(x, awkward1.layout.FillableArray):
            return x.snapshot()
        elif isinstance(x, (awkward1.layout.Content, awkward1.layout.Record)):
            return x
        elif isinstance(x, numpy.ndarray):
            if issubclass(x.dtype.type, numpy.number):
                return awkward1.highlevel.Array(x).layout
            else:
                raise ValueError("numpy.ndarray with {0} cannot be used in {1}".format(repr(x.dtype), ufunc))
        elif isinstance(x, Iterable):
            return unwrap(numpy.array(x))
        else:
            return x

    def checklength(inputs):
        length = len(inputs[0])
        for x in inputs[1:]:
            if len(x) != length:
                raise ValueError("cannot broadcast {0} of length {1} with {2} of length {3}".format(type(inputs[0]).__name__, length, type(x).__name__, len(x)))

    def apply(inputs):
        # handle implicit right-broadcasting (i.e. NumPy-like)
        if any(isinstance(x, awkward1._util.listtypes) for x in inputs):
            maxdepth = max(x.purelist_depth for x in inputs if isinstance(x, awkward1.layout.Content))
            if maxdepth > 0 and all(x.purelist_isregular for x in inputs if isinstance(x, awkward1.layout.Content)):
                nextinputs = []
                for x in inputs:
                    if isinstance(x, awkward1.layout.Content):
                        while x.purelist_depth < maxdepth:
                            x = awkward1.layout.RegularArray(x, 1)
                    nextinputs.append(x)
                if any(x is not y for x, y in zip(inputs, nextinputs)):
                    return apply(nextinputs)

        # now all lengths must agree
        checklength([x for x in inputs if isinstance(x, awkward1.layout.Content)])

        signature = (ufunc,) + tuple(x.parameters.get("__class__") if isinstance(x, awkward1.layout.Content) else type(x) for x in inputs)
        custom = awkward1._util.regular_functions(functions).get(signature)

        # the rest of this is one switch statement
        if custom is not None:
            return custom(*inputs, **kwargs)

        elif any(isinstance(x, awkward1._util.unknowntypes) for x in inputs):
            return apply([x if not isinstance(x, awkward1._util.unknowntypes) else awkward1.layout.NumpyArray(numpy.array([], dtype=numpy.int64)) for x in inputs])

        elif any(isinstance(x, awkward1.layout.NumpyArray) and x.ndim > 1 for x in inputs):
            return apply([x if not (isinstance(x, awkward1.layout.NumpyArray) and x.ndim > 1) else x.regularize_shape() for x in inputs])

        elif any(isinstance(x, awkward1._util.indexedtypes) for x in inputs):
            return apply([x if not isinstance(x, awkward1._util.indexedtypes) else x.project() for x in inputs])

        elif any(isinstance(x, awkward1._util.uniontypes) for x in inputs):
            tagslist = []
            length = None
            for x in inputs:
                if isinstance(x, awkward1._util.uniontypes):
                    tagslist.append(numpy.asarray(x.tags))
                    if length is None:
                        length = len(tagslist[-1])
                    elif length != len(tagslist[-1]):
                        raise ValueError("cannot broadcast UnionArray of length {0} with UnionArray of length {1}".format(length, len(tagslist[-1])))

            combos = numpy.stack(tagslist, axis=-1)
            combos = combos.view([(str(i), combos.dtype) for i in range(len(tagslist))]).reshape(length)

            tags = numpy.empty(length, dtype=numpy.int8)
            index = numpy.empty(length, dtype=numpy.int64)
            contents = []
            for tag, combo in enumerate(numpy.unique(combos)):
                mask = (combos == combo)
                tags[mask] = tag
                index[mask] = numpy.arange(numpy.count_nonzero(mask))
                nextinputs = []
                for i, x in enumerate(inputs):
                    if isinstance(x, awkward1._util.uniontypes):
                        nextinputs.append(x[mask].project(combo[str(i)]))
                    elif isinstance(x, awkward1.layout.Content):
                        nextinputs.append(x[mask])
                    else:
                        nextinputs.append(x)
                contents.append(apply(nextinputs))

            tags = awkward1.layout.Index8(tags)
            index = awkward1.layout.Index64(index)
            return awkward1.layout.UnionArray8_64(tags, index, contents)

        elif any(isinstance(x, awkward1._util.optiontypes) for x in inputs):
            mask = None
            for x in inputs:
                if isinstance(x, (awkward1.layout.IndexedOptionArray32, awkward1.layout.IndexedOptionArray64)):
                    m = numpy.asarray(x.index) < 0
                    if mask is None:
                        mask = m
                    else:
                        numpy.bitwise_or(mask, m, out=mask)

            nextmask = awkward1.layout.Index8(mask.view(numpy.int8))
            index = numpy.full(len(mask), -1, dtype=numpy.int64)
            index[~mask] = numpy.arange(len(mask) - numpy.count_nonzero(mask), dtype=numpy.int64)
            index = awkward1.layout.Index64(index)
            if any(not isinstance(x, awkward1._util.optiontypes) for x in inputs):
                nextindex = numpy.arange(len(mask), dtype=numpy.int64)
                nextindex[mask] = -1
                nextindex = awkward1.layout.Index64(nextindex)

            nextinputs = []
            for x in inputs:
                if isinstance(x, awkward1._util.optiontypes):
                    nextinputs.append(x.project(nextmask))
                else:
                    nextinputs.append(awkward1.layout.IndexedOptionArray64(nextindex, x).project(nextmask))

            return awkward1.layout.IndexedOptionArray64(index, apply(nextinputs))

        elif any(isinstance(x, awkward1._util.listtypes) for x in inputs):
            if all(isinstance(x, awkward1.layout.RegularArray) or not isinstance(x, awkward1._util.listtypes) for x in inputs):
                maxsize = max([x.size for x in inputs if isinstance(x, awkward1.layout.RegularArray)])
                for x in inputs:
                    if isinstance(x, awkward1.layout.RegularArray):
                        if maxsize > 1 and x.size == 1:
                            tmpindex = awkward1.layout.Index64(numpy.repeat(numpy.arange(len(x), dtype=numpy.int64), maxsize))
                nextinputs = []
                for x in inputs:
                    if isinstance(x, awkward1.layout.RegularArray):
                        if maxsize > 1 and x.size == 1:
                            nextinputs.append(awkward1.layout.IndexedArray64(tmpindex, x.content).project())
                        elif x.size == maxsize:
                            nextinputs.append(x.content)
                        else:
                            raise ValueError("cannot broadcast RegularArray of size {0} with RegularArray of size {1}".format(x.size, maxsize))
                    else:
                        nextinputs.append(x)
                return awkward1.layout.RegularArray(apply(nextinputs), maxsize)

            else:
                for x in inputs:
                    if isinstance(x, awkward1._util.listtypes) and not isinstance(x, awkward1.layout.RegularArray):
                        first = x
                        break
                offsets = first.compact_offsets64()
                nextinputs = []
                for x in inputs:
                    if isinstance(x, awkward1._util.listtypes):
                        nextinputs.append(x.broadcast_tooffsets64(offsets).content)
                    # handle implicit left-broadcasting (unlike NumPy)
                    elif isinstance(x, awkward1.layout.Content):
                        nextinputs.append(awkward1.layout.RegularArray(x, 1).broadcast_tooffsets64(offsets).content)
                    else:
                        nextinputs.append(x)
                return awkward1.layout.ListOffsetArray64(offsets, apply(nextinputs))

        elif any(isinstance(x, awkward1._util.recordtypes) for x in inputs):
            keys = None
            length = None
            istuple = True
            for x in inputs:
                if isinstance(x, awkward1._util.recordtypes):
                    if keys is None:
                        keys = x.keys()
                    elif set(keys) != set(x.keys()):
                        raise ValueError("cannot broadcast records because keys don't match:\n    {0}\n    {1}".format(", ".join(sorted(keys)), ", ".join(sorted(x.keys()))))
                    if length is None:
                        length = len(x)
                    elif length != len(x):
                        raise ValueError("cannot broadcast RecordArray of length {0} with RecordArray of length {1}".format(length, len(x)))
                    if not x.istuple:
                        istuple = False

            if len(keys) == 0:
                return awkward1.layout.RecordArray(length, istuple)
            else:
                contents = []
                for key in keys:
                    contents.append(apply([x if not isinstance(x, awkward1._util.recordtypes) else x[key] for x in inputs]))
                return awkward1.layout.RecordArray(contents, keys)

        else:
            assert all(isinstance(x, awkward1.layout.NumpyArray) or not isinstance(x, awkward1.layout.Content) for x in inputs)
            result = getattr(ufunc, method)(*inputs, **kwargs)
            return awkward1.layout.NumpyArray(result)

    isscalar = []

    def pack(inputs):
        maxlen = -1
        for x in inputs:
            if isinstance(x, awkward1.layout.Content):
                maxlen = max(maxlen, len(x))
        if maxlen < 0:
            maxlen = 1
        nextinputs = []
        for x in inputs:
            if isinstance(x, awkward1.layout.Record):
                index = numpy.full(maxlen, x.at, dtype=numpy.int64)
                nextinputs.append(awkward1.layout.RegularArray(x.array[index], maxlen))
                isscalar.append(True)
            elif isinstance(x, awkward1.layout.Content):
                nextinputs.append(awkward1.layout.RegularArray(x, len(x)))
                isscalar.append(False)
            else:
                nextinputs.append(x)
                isscalar.append(True)
        return nextinputs

    def unpack(x):
        if all(isscalar):
            if len(x) == 0:
                return x.getitem_nothing().getitem_nothing()
            else:
                return x[0][0]
        else:
            if len(x) == 0:
                return x.getitem_nothing()
            else:
                return x[0]

    return awkward1._util.wrap(unpack(apply(pack([unwrap(x) for x in inputs]))), classes, functions)

try:
    NDArrayOperatorsMixin = numpy.lib.mixins.NDArrayOperatorsMixin

except AttributeError:
    from numpy.core import umath as um

    def _disables_array_ufunc(obj):
        """True when __array_ufunc__ is set to None."""
        try:
            return obj.__array_ufunc__ is None
        except AttributeError:
            return False

    def _binary_method(ufunc, name):
        """Implement a forward binary method with a ufunc, e.g., __add__."""
        def func(self, other):
            if _disables_array_ufunc(other):
                return NotImplemented
            return ufunc(self, other)
        func.__name__ = "__{}__".format(name)
        return func

    def _reflected_binary_method(ufunc, name):
        """Implement a reflected binary method with a ufunc, e.g., __radd__."""
        def func(self, other):
            if _disables_array_ufunc(other):
                return NotImplemented
            return ufunc(other, self)
        func.__name__ = "__r{}__".format(name)
        return func

    def _inplace_binary_method(ufunc, name):
        """Implement an in-place binary method with a ufunc, e.g., __iadd__."""
        def func(self, other):
            return ufunc(self, other, out=(self,))
        func.__name__ = "__i{}__".format(name)
        return func

    def _numeric_methods(ufunc, name):
        """Implement forward, reflected and inplace binary methods with a ufunc."""
        return (_binary_method(ufunc, name),
                _reflected_binary_method(ufunc, name),
                _inplace_binary_method(ufunc, name))

    def _unary_method(ufunc, name):
        """Implement a unary special method with a ufunc."""
        def func(self):
            return ufunc(self)
        func.__name__ = "__{}__".format(name)
        return func

    class NDArrayOperatorsMixin(object):
        __lt__ = _binary_method(um.less, "lt")
        __le__ = _binary_method(um.less_equal, "le")
        __eq__ = _binary_method(um.equal, "eq")
        __ne__ = _binary_method(um.not_equal, "ne")
        __gt__ = _binary_method(um.greater, "gt")
        __ge__ = _binary_method(um.greater_equal, "ge")

        # numeric methods
        __add__, __radd__, __iadd__ = _numeric_methods(um.add, "add")
        __sub__, __rsub__, __isub__ = _numeric_methods(um.subtract, "sub")
        __mul__, __rmul__, __imul__ = _numeric_methods(um.multiply, "mul")
        if sys.version_info.major < 3:
            # Python 3 uses only __truediv__ and __floordiv__
            __div__, __rdiv__, __idiv__ = _numeric_methods(um.divide, "div")
        __truediv__, __rtruediv__, __itruediv__ = _numeric_methods(
            um.true_divide, "truediv")
        __floordiv__, __rfloordiv__, __ifloordiv__ = _numeric_methods(
            um.floor_divide, "floordiv")
        __mod__, __rmod__, __imod__ = _numeric_methods(um.remainder, "mod")
        if hasattr(um, "divmod"):
            __divmod__ = _binary_method(um.divmod, "divmod")
            __rdivmod__ = _reflected_binary_method(um.divmod, "divmod")
        # __idivmod__ does not exist
        # TODO: handle the optional third argument for __pow__?
        __pow__, __rpow__, __ipow__ = _numeric_methods(um.power, "pow")
        __lshift__, __rlshift__, __ilshift__ = _numeric_methods(
            um.left_shift, "lshift")
        __rshift__, __rrshift__, __irshift__ = _numeric_methods(
            um.right_shift, "rshift")
        __and__, __rand__, __iand__ = _numeric_methods(um.bitwise_and, "and")
        __xor__, __rxor__, __ixor__ = _numeric_methods(um.bitwise_xor, "xor")
        __or__, __ror__, __ior__ = _numeric_methods(um.bitwise_or, "or")

        # unary methods
        __neg__ = _unary_method(um.negative, "neg")
        if hasattr(um, "positive"):
            __pos__ = _unary_method(um.positive, "pos")
            __abs__ = _unary_method(um.absolute, "abs")
            __invert__ = _unary_method(um.invert, "invert")



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import gc

import pytest
import numpy

import awkward1

py27 = 2 if sys.version_info[0] < 3 else 1

def test_refcount1():
    i = numpy.arange(12, dtype="i4").reshape(3, 4)
    assert sys.getrefcount(i) == 2

    i2 = awkward1.layout.Identities32(awkward1.layout.Identities32.newref(), [(0, "hey"), (1, "there")], i)
    assert (sys.getrefcount(i), sys.getrefcount(i2)) == (3, 2)

    tmp = numpy.asarray(i2)
    assert tmp.tolist() == [[0,  1,  2,  3],
                            [4,  5,  6,  7],
                            [8,  9, 10, 11]]

    assert (sys.getrefcount(i), sys.getrefcount(i2)) == (3, 2 + 1*py27)

    del tmp
    assert (sys.getrefcount(i), sys.getrefcount(i2)) == (3, 2)

    tmp2 = i2.array
    assert tmp2.tolist() == [[0,  1,  2,  3],
                             [4,  5,  6,  7],
                             [8,  9, 10, 11]]

    assert (sys.getrefcount(i), sys.getrefcount(i2)) == (3, 2 + 1*py27)

    del tmp2
    assert (sys.getrefcount(i), sys.getrefcount(i2)) == (3, 2)

    del i2
    assert sys.getrefcount(i) == 2

def test_refcount2():
    i = numpy.arange(6, dtype="i4").reshape(3, 2)
    i2 = awkward1.layout.Identities32(awkward1.layout.Identities32.newref(), [], i)
    x = numpy.arange(12).reshape(3, 4)
    x2 = awkward1.layout.NumpyArray(x)
    x2.identities = i2
    del i
    del i2
    del x
    i3 = x2.identities
    del x2
    gc.collect()
    assert numpy.asarray(i3).tolist() == [[0, 1], [2, 3], [4, 5]]
    del i3
    gc.collect()

def test_refcount3():
    i = numpy.arange(6, dtype="i4").reshape(3, 2)
    i2 = awkward1.layout.Identities32(awkward1.layout.Identities32.newref(), [], i)
    x = numpy.arange(12).reshape(3, 4)
    x2 = awkward1.layout.NumpyArray(x)
    x2.identities = i2
    del i2
    assert sys.getrefcount(i) == 3
    x2.identities = None
    assert sys.getrefcount(i) == 2

def test_numpyarray_setidentities():
    x = numpy.arange(160).reshape(40, 4)
    x2 = awkward1.layout.NumpyArray(x)
    x2.setidentities()
    assert numpy.asarray(x2.identities).tolist() == numpy.arange(40).reshape(40, 1).tolist()

def test_listoffsetarray_setidentities():
    content = awkward1.layout.NumpyArray(numpy.arange(10))
    offsets = awkward1.layout.Index32(numpy.array([0, 3, 3, 5, 10], dtype="i4"))
    jagged = awkward1.layout.ListOffsetArray32(offsets, content)
    jagged.setidentities()
    assert numpy.asarray(jagged.identities).tolist() == [[0], [1], [2], [3]]
    assert numpy.asarray(jagged.content.identities).tolist() == [[0, 0], [0, 1], [0, 2], [2, 0], [2, 1], [3, 0], [3, 1], [3, 2], [3, 3], [3, 4]]

    assert numpy.asarray(jagged.content[3:7].identities).tolist() == [[2, 0], [2, 1], [3, 0], [3, 1]]
    assert numpy.asarray(jagged[0].identities).tolist() == [[0, 0], [0, 1], [0, 2]]
    assert numpy.asarray(jagged[1].identities).tolist() == []
    assert numpy.asarray(jagged[2].identities).tolist() == [[2, 0], [2, 1]]
    assert numpy.asarray(jagged[3].identities).tolist() == [[3, 0], [3, 1], [3, 2], [3, 3], [3, 4]]
    assert numpy.asarray(jagged[1:3].identities).tolist() == [[1], [2]]

def test_setidentities_none():
    offsets = awkward1.layout.Index32(numpy.array([0, 2, 2, 3], "i4"))
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3]))
    array = awkward1.layout.ListOffsetArray32(offsets, content)
    assert array.identities is None
    assert array.content.identities is None
    array.identities = None
    assert array.identities is None
    assert array.content.identities is None
    repr(array)
    array.setidentities()
    repr(array)
    assert array.identities is not None
    assert array.content.identities is not None
    array.identities = None
    assert array.identities is None
    assert array.content.identities is None

def test_setidentities_constructor():
    offsets = awkward1.layout.Index32(numpy.array([0, 2, 2, 3], "i4"))
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3]), identities=awkward1.layout.Identities32(awkward1.layout.Identities32.newref(), [], numpy.array([[0, 0], [0, 1], [2, 0]], dtype="i4")))
    array = awkward1.layout.ListOffsetArray32(offsets, content, identities=awkward1.layout.Identities32(awkward1.layout.Identities32.newref(), [], numpy.array([[0], [1], [2]], dtype="i4")))
    assert numpy.asarray(array.identities).tolist() == [[0], [1], [2]]
    assert numpy.asarray(array.content.identities).tolist() == [[0, 0], [0, 1], [2, 0]]



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import os
import json

import pytest
import numpy

import awkward1

def test_fromstring():
    a = awkward1.fromjson("[[1.1, 2.2, 3], [], [4, 5.5]]")
    assert awkward1.tolist(a) == [[1.1, 2.2, 3.0], [], [4.0, 5.5]]

    with pytest.raises(ValueError):
        awkward1.fromjson("[[1.1, 2.2, 3], [blah], [4, 5.5]]")

def test_fromfile(tmp_path):
    with open(os.path.join(str(tmp_path), "tmp1.json"), "w") as f:
        f.write("[[1.1, 2.2, 3], [], [4, 5.5]]")

    a = awkward1.fromjson(os.path.join(str(tmp_path), "tmp1.json"))
    assert awkward1.tolist(a) == [[1.1, 2.2, 3.0], [], [4.0, 5.5]]

    with pytest.raises(ValueError):
        awkward1.fromjson("nonexistent.json")

    with open(os.path.join(str(tmp_path), "tmp2.json"), "w") as f:
        f.write("[[1.1, 2.2, 3], []], [4, 5.5]]")

    with pytest.raises(ValueError):
        awkward1.fromjson(os.path.join(str(tmp_path), "tmp2.json"))

def test_tostring():
    content = awkward1.layout.NumpyArray(numpy.arange(2*3*5*7).reshape(-1, 7))
    offsetsA = numpy.arange(0, 2*3*5 + 5, 5)
    offsetsB = numpy.arange(0, 2*3 + 3, 3)
    startsA, stopsA = offsetsA[:-1], offsetsA[1:]
    startsB, stopsB = offsetsB[:-1], offsetsB[1:]

    listoffsetarrayA32 = awkward1.layout.ListOffsetArray32(awkward1.layout.Index32(offsetsA), content)
    listarrayA32 = awkward1.layout.ListArray32(awkward1.layout.Index32(startsA), awkward1.layout.Index32(stopsA), content)
    modelA = numpy.arange(2*3*5*7).reshape(2*3, 5, 7)

    listoffsetarrayB32 = awkward1.layout.ListOffsetArray32(awkward1.layout.Index32(offsetsB), listoffsetarrayA32)
    listarrayB32 = awkward1.layout.ListArray32(awkward1.layout.Index32(startsB), awkward1.layout.Index32(stopsB), listarrayA32)
    modelB = numpy.arange(2*3*5*7).reshape(2, 3, 5, 7)

    assert content.tojson() == json.dumps(awkward1.tolist(content), separators=(",", ":"))
    assert listoffsetarrayA32.tojson() == json.dumps(modelA.tolist(), separators=(",", ":"))
    assert listoffsetarrayB32.tojson() == json.dumps(modelB.tolist(), separators=(",", ":"))
    awkward1.tojson(awkward1.fromjson("[[1.1,2.2,3],[],[4,5.5]]")) == "[[1.1,2.2,3],[],[4,5.5]]"

def test_tofile(tmp_path):
    awkward1.tojson(awkward1.fromjson("[[1.1,2.2,3],[],[4,5.5]]"), os.path.join(str(tmp_path), "tmp1.json"))

    with open(os.path.join(str(tmp_path), "tmp1.json"), "r") as f:
        f.read() == "[[1.1,2.2,3],[],[4,5.5]]"

def test_root_nestedvector():
    # fromcounts([3, 2], fromcounts([1, 0, 2, 2, 1], [123, 99, 123, 99, 123, 123]))
    # <JaggedArray [[[123] [] [99 123]] [[99 123] [123]]]>

    # outer offsets: [0, 3, 5]
    # inner offsets: [0, 1, 1, 3, 5, 6]

    byteoffsets = awkward1.layout.Index64(numpy.array([0, 28, 52], dtype=numpy.int64))
    rawdata = awkward1.layout.NumpyArray(numpy.array([
        0, 0, 0, 3, 0, 0, 0, 1, 0, 0, 0, 123, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 99, 0, 0, 0, 123,
        0, 0, 0, 2, 0, 0, 0, 2, 0, 0, 0, 99, 0, 0, 0, 123, 0, 0, 0, 1, 0, 0, 0, 123
        ], dtype=numpy.uint8))

    result = awkward1.layout.fromroot_nestedvector(byteoffsets, rawdata, 2, numpy.dtype(">i").itemsize, ">i")
    assert numpy.asarray(result.offsets).tolist() == [0, 3, 5]
    assert numpy.asarray(result.content.offsets).tolist() == [0, 1, 1, 3, 5, 6]
    assert numpy.asarray(result.content.content).tolist() == [123, 99, 123, 99, 123, 123]
    assert awkward1.tolist(result) == [[[123], [], [99, 123]], [[99, 123], [123]]]

def test_fromiter():
    assert awkward1.tolist(awkward1.fromiter([True, True, False, False, True])) == [True, True, False, False, True]
    assert awkward1.tolist(awkward1.fromiter([5, 4, 3, 2, 1])) == [5, 4, 3, 2, 1]
    assert awkward1.tolist(awkward1.fromiter([5, 4, 3.14, 2.22, 1.23])) == [5.0, 4.0, 3.14, 2.22, 1.23]
    assert awkward1.tolist(awkward1.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]])) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
    assert awkward1.tolist(awkward1.fromiter([[[1.1, 2.2, 3.3], []], [[4.4, 5.5]], [], [[6.6], [7.7, 8.8, 9.9]]])) == [[[1.1, 2.2, 3.3], []], [[4.4, 5.5]], [], [[6.6], [7.7, 8.8, 9.9]]]

def test_numpy():
    a = awkward1.layout.NumpyArray(numpy.arange(2*3*5).reshape(2, 3, 5))
    assert awkward1.tolist(a) == [[[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]], [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]]]
    assert awkward1.tojson(a) == "[[[0,1,2,3,4],[5,6,7,8,9],[10,11,12,13,14]],[[15,16,17,18,19],[20,21,22,23,24],[25,26,27,28,29]]]"

    b = awkward1.layout.NumpyArray(numpy.array([[[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]], [[10.1, 20.2, 30.3], [40.4, 50.5, 60.6]]]))
    assert awkward1.tojson(b) == "[[[1.1,2.2,3.3],[4.4,5.5,6.6]],[[10.1,20.2,30.3],[40.4,50.5,60.6]]]"

    c = awkward1.layout.NumpyArray(numpy.array([[True, False, True], [False, False, True]]))
    assert awkward1.tojson(c) == "[[true,false,true],[false,false,true]]"



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys

import pytest
import numpy

import awkward1

def test_typeempty():
    assert numpy.asarray(awkward1.layout.PrimitiveType("bool").empty()).dtype == numpy.dtype("bool")
    assert numpy.asarray(awkward1.layout.PrimitiveType("int8").empty()).dtype == numpy.dtype("int8")
    assert numpy.asarray(awkward1.layout.PrimitiveType("int16").empty()).dtype == numpy.dtype("int16")
    assert numpy.asarray(awkward1.layout.PrimitiveType("int32").empty()).dtype == numpy.dtype("int32")
    assert numpy.asarray(awkward1.layout.PrimitiveType("int64").empty()).dtype == numpy.dtype("int64")
    assert numpy.asarray(awkward1.layout.PrimitiveType("uint8").empty()).dtype == numpy.dtype("uint8")
    assert numpy.asarray(awkward1.layout.PrimitiveType("uint16").empty()).dtype == numpy.dtype("uint16")
    assert numpy.asarray(awkward1.layout.PrimitiveType("uint32").empty()).dtype == numpy.dtype("uint32")
    assert numpy.asarray(awkward1.layout.PrimitiveType("uint64").empty()).dtype == numpy.dtype("uint64")
    assert numpy.asarray(awkward1.layout.PrimitiveType("float32").empty()).dtype == numpy.dtype("float32")
    assert numpy.asarray(awkward1.layout.PrimitiveType("float64").empty()).dtype == numpy.dtype("float64")
    assert type(awkward1.layout.UnknownType().empty()) is awkward1.layout.EmptyArray
    assert type(awkward1.layout.ArrayType(awkward1.layout.UnknownType(), 0).empty()) is awkward1.layout.EmptyArray
    assert type(awkward1.layout.RegularType(awkward1.layout.UnknownType(), 5).empty()) is awkward1.layout.RegularArray
    assert type(awkward1.layout.ListType(awkward1.layout.UnknownType()).empty()) is awkward1.layout.ListOffsetArray64
    array = awkward1.layout.RecordType({"one": awkward1.layout.PrimitiveType("float64"), "two": awkward1.layout.ListType(awkward1.layout.PrimitiveType("float64"))}).empty()
    assert type(array) is awkward1.layout.RecordArray
    assert type(array["one"]) is awkward1.layout.NumpyArray
    assert numpy.asarray(array["one"]).dtype == numpy.dtype("float64")
    assert type(array["two"]) is awkward1.layout.ListOffsetArray64

def test_astype():
    empty = awkward1.layout.EmptyArray()
    assert numpy.asarray(empty.astype(awkward1.layout.PrimitiveType("bool"))).dtype == numpy.dtype("bool")
    assert numpy.asarray(empty.astype(awkward1.layout.PrimitiveType("uint8"))).dtype == numpy.dtype("uint8")
    assert numpy.asarray(empty.astype(awkward1.layout.PrimitiveType("float64"))).dtype == numpy.dtype("float64")
    assert type(empty.astype(awkward1.layout.ListType(awkward1.layout.PrimitiveType("float64")))) is awkward1.layout.ListOffsetArray64



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import os
import ctypes
import platform
import glob
import xml.etree.ElementTree

import numba.typing.ctypes_utils

if platform.system() == "Windows":
    libname = "awkward-cpu-kernels.dll"
elif platform.system() == "Darwin":
    libname = "libawkward-cpu-kernels.dylib"
else:
    libname = "libawkward-cpu-kernels.so"

libpath = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), libname)

kernels = ctypes.cdll.LoadLibrary(libpath)

class ErrorType(numba.types.Type):
    def __init__(self):
        super(ErrorType, self).__init__(name="ak::ErrorType")

class Error(ctypes.Structure):
    _fields_ = [("str", ctypes.c_char_p),
                ("identity", ctypes.c_int64),
                ("attempt", ctypes.c_int64),
                ("extra", ctypes.c_int64)]
    numbatpe = ErrorType()

@numba.extending.register_model(ErrorType)
class ErrorModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [("str", numba.intp),
                   ("identity", numba.int64),
                   ("attempt", numba.int64),
                   ("extra", numba.int64)]
        super(ErrorModel, self).__init__(dmm, fe_type, members)

h2ctypes = {
    "bool": ctypes.c_uint8,
    "bool *": ctypes.POINTER(ctypes.c_uint8),
    "int8_t *": ctypes.POINTER(ctypes.c_int8),
    "const int8_t *": ctypes.POINTER(ctypes.c_int8),
    "uint8_t *": ctypes.POINTER(ctypes.c_uint8),
    "const uint8_t *": ctypes.POINTER(ctypes.c_uint8),
    "int32_t": ctypes.c_int32,
    "int32_t *": ctypes.POINTER(ctypes.c_int32),
    "const int32_t *": ctypes.POINTER(ctypes.c_int32),
    "uint32_t": ctypes.c_uint32,
    "uint32_t *": ctypes.POINTER(ctypes.c_uint32),
    "const uint32_t *": ctypes.POINTER(ctypes.c_uint32),
    "int64_t": ctypes.c_int64,
    "int64_t *": ctypes.POINTER(ctypes.c_int64),
    "const int64_t *": ctypes.POINTER(ctypes.c_int64),
    "Error": Error,
    "ERROR": Error,
    "void": None,
    }

for hfile in glob.glob(os.path.join(os.path.join(os.path.dirname(os.path.dirname(__file__)), "signatures"), "*_8cpp.xml")):
    xfile = xml.etree.ElementTree.parse(hfile)
    if xfile.find("./compounddef/location").attrib["file"].startswith("src/cpu-kernels"):
        for xfcn in xfile.findall(".//memberdef[@kind='function']"):
            name = xfcn.find("./name").text
            if hasattr(kernels, name) and name not in ("success", "failure"):
                rettype = xfcn.find("./type").text
                params = [(x.find("./declname").text, x.find("./type").text) for x in xfcn.findall("./param")]
                getattr(kernels, name).name = name
                getattr(kernels, name).argtypes = [h2ctypes[t] for n, t in params]
                if rettype == "Error" or rettype == "ERROR":
                    getattr(kernels, name).restype = None
                    tmp = numba.typing.ctypes_utils.make_function_type(getattr(kernels, name))
                    getattr(kernels, name).numbatpe = numba.types.functions.ExternalFunctionPointer(Error.numbatpe(*tmp.sig.args), tmp.get_pointer, cconv=tmp.cconv)
                    getattr(kernels, name).restype = h2ctypes[rettype]
                else:
                    getattr(kernels, name).restype = h2ctypes[rettype]
                    getattr(kernels, name).numbatpe = numba.typing.ctypes_utils.make_function_type(getattr(kernels, name))



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys

import pytest
import numpy

import awkward1

def test_mixing_lists_and_none():
    def add(a, b):
        outer = []
        for ai, bi in zip(a, b):
            if ai is None or bi is None:
                outer.append(None)
            else:
                inner = []
                for aj, bj in zip(ai, bi):
                    if aj is None or bj is None:
                        inner.append(None)
                    else:
                        inner.append(aj + bj)
                outer.append(inner)
        return outer

    a00 = awkward1.Array([[1.1,  2.2, 3.3],   [], [4.4, 5.5], [ 6.6, 7.7, 8.8, 9.9]])
    a01 = awkward1.Array([[1.1, None, 3.3],   [], [4.4, 5.5], [ 6.6, 7.7, 8.8, 9.9]])
    a02 = awkward1.Array([[1.1, None, 3.3],   [], [4.4, 5.5], [None, 7.7, 8.8, 9.9]])
    a10 = awkward1.Array([[1.1,  2.2, 3.3],   [],       None, [ 6.6, 7.7, 8.8, 9.9]])
    a11 = awkward1.Array([[1.1, None, 3.3],   [],       None, [ 6.6, 7.7, 8.8, 9.9]])
    a12 = awkward1.Array([[1.1, None, 3.3],   [],       None, [None, 7.7, 8.8, 9.9]])
    a20 = awkward1.Array([[1.1,  2.2, 3.3], None,       None, [ 6.6, 7.7, 8.8, 9.9]])
    a21 = awkward1.Array([[1.1, None, 3.3], None,       None, [ 6.6, 7.7, 8.8, 9.9]])
    a22 = awkward1.Array([[1.1, None, 3.3], None,       None, [None, 7.7, 8.8, 9.9]])

    b00 = awkward1.Array([[100,  200, 300],   [], [400, 500], [ 600, 700, 800, 900]])
    b01 = awkward1.Array([[100, None, 300],   [], [400, 500], [ 600, 700, 800, 900]])
    b02 = awkward1.Array([[100, None, 300],   [], [400, 500], [None, 700, 800, 900]])
    b10 = awkward1.Array([[100,  200, 300],   [],       None, [ 600, 700, 800, 900]])
    b11 = awkward1.Array([[100, None, 300],   [],       None, [ 600, 700, 800, 900]])
    b12 = awkward1.Array([[100, None, 300],   [],       None, [None, 700, 800, 900]])
    b20 = awkward1.Array([[100,  200, 300], None,       None, [ 600, 700, 800, 900]])
    b21 = awkward1.Array([[100, None, 300], None,       None, [ 600, 700, 800, 900]])
    b22 = awkward1.Array([[100, None, 300], None,       None, [None, 700, 800, 900]])

    for a in (a00, a01, a02, a10, a11, a12, a20, a21, a22):
        for b in (b00, b01, b02, b10, b11, b12, b20, b21, b22):
            assert awkward1.tolist(a + b) == add(a, b)

def test_explicit_broadcasting():
    nparray = numpy.arange(2*3*5).reshape(2, 3, 5)
    lsarray = awkward1.Array(nparray.tolist())
    rgarray = awkward1.Array(nparray)

    # explicit left-broadcasting
    assert awkward1.tolist(rgarray + numpy.array([[[100]], [[200]]])) == awkward1.tolist(nparray + numpy.array([[[100]], [[200]]]))
    assert awkward1.tolist(lsarray + numpy.array([[[100]], [[200]]])) == awkward1.tolist(nparray + numpy.array([[[100]], [[200]]]))
    assert awkward1.tolist(numpy.array([[[100]], [[200]]]) + rgarray) == awkward1.tolist(numpy.array([[[100]], [[200]]]) + nparray)
    assert awkward1.tolist(numpy.array([[[100]], [[200]]]) + lsarray) == awkward1.tolist(numpy.array([[[100]], [[200]]]) + nparray)

    # explicit right-broadcasting
    assert awkward1.tolist(rgarray + numpy.array([[[100, 200, 300, 400, 500]]])) == awkward1.tolist(nparray + numpy.array([[[100, 200, 300, 400, 500]]]))
    assert awkward1.tolist(lsarray + numpy.array([[[100, 200, 300, 400, 500]]])) == awkward1.tolist(nparray + numpy.array([[[100, 200, 300, 400, 500]]]))
    assert awkward1.tolist(numpy.array([[[100, 200, 300, 400, 500]]]) + rgarray) == awkward1.tolist(numpy.array([[[100, 200, 300, 400, 500]]]) + nparray)
    assert awkward1.tolist(numpy.array([[[100, 200, 300, 400, 500]]]) + lsarray) == awkward1.tolist(numpy.array([[[100, 200, 300, 400, 500]]]) + nparray)

def test_implicit_broadcasting():
    nparray = numpy.arange(2*3*5).reshape(2, 3, 5)
    lsarray = awkward1.Array(nparray.tolist())
    rgarray = awkward1.Array(nparray)

    assert awkward1.tolist(rgarray + numpy.array([100, 200, 300, 400, 500])) == awkward1.tolist(nparray + numpy.array([100, 200, 300, 400, 500]))
    assert awkward1.tolist(numpy.array([100, 200, 300, 400, 500]) + rgarray) == awkward1.tolist(numpy.array([100, 200, 300, 400, 500]) + nparray)

    assert awkward1.tolist(lsarray + numpy.array([100, 200])) == awkward1.tolist(nparray + numpy.array([[[100]], [[200]]]))
    assert awkward1.tolist(numpy.array([100, 200]) + lsarray) == awkward1.tolist(numpy.array([[[100]], [[200]]]) + nparray)

def test_records():
    array = awkward1.Array([[{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}], [], [{"x": 4, "y": 4.4}, {"x": 5, "y": 5.5}], [{"x": 100, "y": 200}]])
    record = array[-1, -1]

    assert awkward1.tolist(array[0] + record) == [{"x": 101, "y": 201.1}, {"x": 102, "y": 202.2}, {"x": 103, "y": 203.3}]

    assert awkward1.tolist(array + record) == [[{"x": 101, "y": 201.1}, {"x": 102, "y": 202.2}, {"x": 103, "y": 203.3}], [], [{"x": 104, "y": 204.4}, {"x": 105, "y": 205.5}], [{"x": 200, "y": 400.0}]]

    assert awkward1.tolist(record + record) == {"x": 200, "y": 400}

def test_tonumpy():
    assert numpy.array_equal(awkward1.tonumpy(awkward1.Array([1.1, 2.2, 3.3, 4.4, 5.5])), numpy.array([1.1, 2.2, 3.3, 4.4, 5.5]))
    assert numpy.array_equal(awkward1.tonumpy(awkward1.Array(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5]))), numpy.array([1.1, 2.2, 3.3, 4.4, 5.5]))
    assert numpy.array_equal(awkward1.tonumpy(awkward1.Array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6]])), numpy.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6]]))
    assert numpy.array_equal(awkward1.tonumpy(awkward1.Array(numpy.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6]]))), numpy.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6]]))
    assert numpy.array_equal(awkward1.tonumpy(awkward1.Array(["one", "two", "three"])), numpy.array(["one", "two", "three"]))
    assert numpy.array_equal(awkward1.tonumpy(awkward1.Array([b"one", b"two", b"three"])), numpy.array([b"one", b"two", b"three"]))
    assert numpy.array_equal(awkward1.tonumpy(awkward1.Array([])), numpy.array([]))

    content0 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5], dtype=numpy.float64))
    content1 = awkward1.layout.NumpyArray(numpy.array([1, 2, 3], dtype=numpy.int64))
    tags = awkward1.layout.Index8(numpy.array([0, 1, 1, 0, 0, 0, 1, 0], dtype=numpy.int8))
    index = awkward1.layout.Index64(numpy.array([0, 0, 1, 1, 2, 3, 2, 4], dtype=numpy.int64))
    array = awkward1.Array(awkward1.layout.UnionArray8_64(tags, index, [content0, content1]))
    assert numpy.array_equal(awkward1.tonumpy(array), numpy.array([1.1, 1, 2, 2.2, 3.3, 4.4, 3, 5.5]))

    assert awkward1.tonumpy(awkward1.Array([1.1, 2.2, None, None, 3.3])).tolist() == [1.1, 2.2, None, None, 3.3]
    assert awkward1.tonumpy(awkward1.Array([[1.1, 2.2], [None, None], [3.3, 4.4]])).tolist() == [[1.1, 2.2], [None, None], [3.3, 4.4]]
    assert awkward1.tonumpy(awkward1.Array([[1.1, 2.2], None, [3.3, 4.4]])).tolist() == [[1.1, 2.2], [None, None], [3.3, 4.4]]

    assert numpy.array_equal(awkward1.tonumpy(awkward1.Array([{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}])), numpy.array([(1, 1.1), (2, 2.2), (3, 3.3)], dtype=[("x", numpy.int64), ("y", numpy.float64)]))
    assert numpy.array_equal(awkward1.tonumpy(awkward1.Array([(1, 1.1), (2, 2.2), (3, 3.3)])), numpy.array([(1, 1.1), (2, 2.2), (3, 3.3)], dtype=[("0", numpy.int64), ("1", numpy.float64)]))

def test_numpy_array():
    assert numpy.array_equal(numpy.asarray(awkward1.Array([1.1, 2.2, 3.3, 4.4, 5.5])), numpy.array([1.1, 2.2, 3.3, 4.4, 5.5]))
    assert numpy.array_equal(numpy.asarray(awkward1.Array(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5]))), numpy.array([1.1, 2.2, 3.3, 4.4, 5.5]))
    assert numpy.array_equal(numpy.asarray(awkward1.Array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6]])), numpy.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6]]))
    assert numpy.array_equal(numpy.asarray(awkward1.Array(numpy.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6]]))), numpy.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6]]))
    assert numpy.array_equal(numpy.asarray(awkward1.Array(["one", "two", "three"])), numpy.array(["one", "two", "three"]))
    assert numpy.array_equal(numpy.asarray(awkward1.Array([b"one", b"two", b"three"])), numpy.array([b"one", b"two", b"three"]))
    assert numpy.array_equal(numpy.asarray(awkward1.Array([])), numpy.array([]))

    content0 = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5], dtype=numpy.float64))
    content1 = awkward1.layout.NumpyArray(numpy.array([1, 2, 3], dtype=numpy.int64))
    tags = awkward1.layout.Index8(numpy.array([0, 1, 1, 0, 0, 0, 1, 0], dtype=numpy.int8))
    index = awkward1.layout.Index64(numpy.array([0, 0, 1, 1, 2, 3, 2, 4], dtype=numpy.int64))
    array = awkward1.Array(awkward1.layout.UnionArray8_64(tags, index, [content0, content1]))
    assert numpy.array_equal(numpy.asarray(array), numpy.array([1.1, 1, 2, 2.2, 3.3, 4.4, 3, 5.5]))

    assert numpy.ma.asarray(awkward1.Array([1.1, 2.2, None, None, 3.3])).tolist() == [1.1, 2.2, None, None, 3.3]
    assert numpy.ma.asarray(awkward1.Array([[1.1, 2.2], [None, None], [3.3, 4.4]])).tolist() == [[1.1, 2.2], [None, None], [3.3, 4.4]]
    assert numpy.ma.asarray(awkward1.Array([[1.1, 2.2], None, [3.3, 4.4]])).tolist() == [[1.1, 2.2], [None, None], [3.3, 4.4]]

    assert numpy.array_equal(numpy.asarray(awkward1.Array([{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}])), numpy.array([(1, 1.1), (2, 2.2), (3, 3.3)], dtype=[("x", numpy.int64), ("y", numpy.float64)]))
    assert numpy.array_equal(numpy.asarray(awkward1.Array([(1, 1.1), (2, 2.2), (3, 3.3)])), numpy.array([(1, 1.1), (2, 2.2), (3, 3.3)], dtype=[("0", numpy.int64), ("1", numpy.float64)]))

def test_where():
    one = awkward1.Array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    two = awkward1.Array([  0, 100, 200, 300, 400, 500, 600, 700, 800, 900])
    condition = awkward1.Array([False, False, False, False, False, True, False, True, False, True])

    assert isinstance(awkward1.where(condition)[0], awkward1.Array)
    assert awkward1.tolist(awkward1.where(condition)[0]) == [5, 7, 9]

    assert awkward1.tolist(awkward1.where(condition, one, two)) == awkward1.tolist(numpy.where(numpy.asarray(condition), numpy.asarray(one), numpy.asarray(two)))

def test_string_equal():
    one = awkward1.Array(["one", "two", "three"])
    two = awkward1.Array(["ONE", "two", "four"])
    assert awkward1.tolist(one == two) == [False, True, False]

def test_size():
    assert awkward1.size(awkward1.Array([1.1, 2.2, 3.3, 4.4, 5.5])) == 5
    assert awkward1.size(awkward1.Array(numpy.arange(2*3*5).reshape(2, 3, 5))) == 30
    assert awkward1.size(awkward1.Array(numpy.arange(2*3*5).reshape(2, 3, 5)), 0) == 2
    assert awkward1.size(awkward1.Array(numpy.arange(2*3*5).reshape(2, 3, 5)), 1) == 3
    assert awkward1.size(awkward1.Array(numpy.arange(2*3*5).reshape(2, 3, 5)), 2) == 5
    assert awkward1.size(awkward1.layout.NumpyArray(numpy.arange(2*3*5).reshape(2, 3, 5))) == 30
    assert awkward1.size(awkward1.layout.NumpyArray(numpy.arange(2*3*5).reshape(2, 3, 5)), 0) == 2
    assert awkward1.size(awkward1.layout.NumpyArray(numpy.arange(2*3*5).reshape(2, 3, 5)), 1) == 3
    assert awkward1.size(awkward1.layout.NumpyArray(numpy.arange(2*3*5).reshape(2, 3, 5)), 2) == 5
    assert numpy.size(numpy.arange(2*3*5).reshape(2, 3, 5)) == 30
    assert numpy.size(numpy.arange(2*3*5).reshape(2, 3, 5), 0) == 2
    assert numpy.size(numpy.arange(2*3*5).reshape(2, 3, 5), 1) == 3
    assert numpy.size(numpy.arange(2*3*5).reshape(2, 3, 5), 2) == 5
    with pytest.raises(ValueError) as err:
        awkward1.size(awkward1.Array(numpy.arange(2*3*5).reshape(2, 3, 5).tolist()))
    assert str(err.value) == "ak.size is ambiguous due to variable-length arrays (try ak.flatten to remove structure or ak.tonumpy to force regularity, if possible)"

plt = pytest.importorskip("matplotlib.pyplot")

def test_inhibit_matplotlib():
    array = awkward1.Array([1.1, 2.2, 3.3, 4.4, 5.5])
    plt.hist(array)
    plt.plot(array)
    plt.plot(array, array)

    array = awkward1.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
    with pytest.raises(ValueError):
        plt.hist(array)
    with pytest.raises(ValueError):
        plt.plot(array)
    with pytest.raises(ValueError):
        plt.plot(array, array)



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import operator

import numpy
import numba

import awkward1.layout
from ..._numba import cpu, util, content

@numba.extending.typeof_impl.register(awkward1.layout.EmptyArray)
def typeof(val, c):
    return EmptyArrayType(numba.typeof(val.identities), util.dict2parameters(val.parameters))

class EmptyArrayType(content.ContentType):
    def __init__(self, identitiestpe, parameters):
        assert isinstance(parameters, tuple)
        super(EmptyArrayType, self).__init__(name="ak::EmptyArrayType(identities={0}, parameters={1})".format(identitiestpe.name, util.parameters2str(parameters)))
        self.identitiestpe = identitiestpe
        self.parameters = parameters

    @property
    def ndim(self):
        return 1

    def getitem_int(self):
        raise ValueError("cannot compile getitem for EmptyArray, which has unknown element type")

    def getitem_range(self):
        return self

    def getitem_str(self):
        raise IndexError("cannot slice EmptyArray with str (Record field name)")

    def getitem_tuple(self, wheretpe):
        if len(wheretpe.types) == 0:
            return self
        elif len(wheretpe.types) == 1 and isinstance(wheretpe.types[0], numba.types.SliceType):
            return self
        else:
            raise ValueError("cannot compile getitem for EmptyArray, which has unknown element type")

    def getitem_next(self, wheretpe, isadvanced):
        if len(wheretpe.types) == 0:
            return self
        else:
            raise ValueError("cannot compile getitem for EmptyArray, which has unknown element type")

    def carry(self):
        return self

    @property
    def lower_len(self):
        return lower_len

    @property
    def lower_getitem_nothing(self):
        return lower_getitem_nothing

    @property
    def lower_getitem_range(self):
        return lower_getitem_range

    @property
    def lower_getitem_next(self):
        return lower_getitem_next

    @property
    def lower_carry(self):
        return lower_carry

@numba.extending.register_model(EmptyArrayType)
class EmptyArrayModel(numba.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = []
        if fe_type.identitiestpe != numba.none:
            members.append(("identities", fe_type.identitiestpe))
        super(EmptyArrayModel, self).__init__(dmm, fe_type, members)

@numba.extending.unbox(EmptyArrayType)
def unbox(tpe, obj, c):
    proxyout = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder)
    if tpe.identitiestpe != numba.none:
        id_obj = c.pyapi.object_getattr_string(obj, "identities")
        proxyout.identities = c.pyapi.to_native_value(tpe.identitiestpe, id_obj).value
        c.pyapi.decref(id_obj)
    is_error = numba.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)

@numba.extending.box(EmptyArrayType)
def box(tpe, val, c):
    EmptyArray_obj = c.pyapi.unserialize(c.pyapi.serialize_object(awkward1.layout.EmptyArray))
    proxyin = numba.cgutils.create_struct_proxy(tpe)(c.context, c.builder, value=val)
    args = []
    if tpe.identitiestpe != numba.none:
        args.append(c.pyapi.from_native_value(tpe.identitiestpe, proxyin.identities, c.env_manager))
    else:
        args.append(c.pyapi.make_none())
    args.append(util.parameters2dict_impl(c, tpe.parameters))
    out = c.pyapi.call_function_objargs(EmptyArray_obj, args)
    for x in args:
        c.pyapi.decref(x)
    c.pyapi.decref(EmptyArray_obj)
    return out

@numba.extending.lower_builtin(len, EmptyArrayType)
def lower_len(context, builder, sig, args):
    return context.get_constant(numba.intp, 0)

def lower_getitem_nothing(context, builder, tpe, val):
    return val

@numba.extending.lower_builtin(operator.getitem, EmptyArrayType, numba.types.slice2_type)
def lower_getitem_range(context, builder, sig, args):
    rettpe, (tpe, wheretpe) = sig.return_type, sig.args
    val, whereval = args
    if context.enable_nrt:
        context.nrt.incref(builder, rettpe, val)
    return val

@numba.extending.lower_builtin(operator.getitem, EmptyArrayType, numba.types.BaseTuple)
def lower_getitem_tuple(context, builder, sig, args):
    rettpe, (tpe, wheretpe) = sig.return_type, sig.args
    val, whereval = args
    if context.enable_nrt:
        context.nrt.incref(builder, rettpe, val)
    return val

def lower_getitem_next(context, builder, arraytpe, wheretpe, arrayval, whereval, advanced):
    if context.enable_nrt:
        context.nrt.incref(builder, arraytpe, arrayval)
    return arrayval

def lower_carry(context, builder, arraytpe, carrytpe, arrayval, carryval):
    if context.enable_nrt:
        context.nrt.incref(builder, arraytpe, arrayval)
    return arrayval

@numba.typing.templates.infer_getattr
class type_methods(numba.typing.templates.AttributeTemplate):
    key = EmptyArrayType

    def generic_resolve(self, tpe, attr):
        if attr == "identities":
            if tpe.identitiestpe == numba.none:
                return numba.optional(identity.IdentitiesType(numba.int32[:, :]))
            else:
                return tpe.identitiestpe

@numba.extending.lower_getattr(EmptyArrayType, "identities")
def lower_identities(context, builder, tpe, val):
    proxyin = numba.cgutils.create_struct_proxy(tpe)(context, builder, value=val)
    if tpe.identitiestpe == numba.none:
        return context.make_optional_none(builder, identity.IdentitiesType(numba.int32[:, :]))
    else:
        if context.enable_nrt:
            context.nrt.incref(builder, tpe.identitiestpe, proxyin.identities)
        return proxyin.identities



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import sys
import itertools

import pytest
import numpy

import awkward1

def test_types():
    t0 = awkward1.layout.UnknownType()
    t1 = awkward1.layout.PrimitiveType("int32")
    t2 = awkward1.layout.OptionType(t1)
    t3 = awkward1.layout.UnionType((t1, awkward1.layout.PrimitiveType("float64")))
    t4 = awkward1.layout.ListType(t1)
    t4b = awkward1.layout.ListType(awkward1.layout.PrimitiveType("int32"))
    t5 = awkward1.layout.ListType(t4)
    t6 = awkward1.layout.OptionType(t4)
    assert repr(t0) == "unknown"
    assert repr(t1) == "int32"
    assert repr(t2) == "?int32"
    assert repr(t3) == "union[int32, float64]"
    assert repr(t4) == "var * int32"
    assert repr(t5) == "var * var * int32"
    assert repr(t6) == "option[var * int32]"
    assert repr(t2.type) == "int32"
    assert t3.numtypes == 2
    assert repr(t3.type(0)) == "int32"
    assert repr(t3.type(1)) == "float64"
    assert [repr(x) for x in t3.types] == ["int32", "float64"]
    assert repr(t4.type) == "int32"
    assert repr(t5.type) == "var * int32"

def test_boolean():
    a = awkward1.layout.FillableArray()
    a.boolean(True)
    a.boolean(True)
    a.boolean(False)
    a.boolean(True)
    assert awkward1.tolist(a.snapshot()) == [True, True, False, True]
    assert awkward1.tolist(a) == [True, True, False, True]
    assert awkward1.tolist(a[1:-1]) == [True, False]

def test_big():
    a = awkward1.layout.FillableArray(initial=90)
    for i in range(2000):
        if i == 200:
            tmp = a.snapshot()
        a.boolean(i % 2 == 0)
    assert awkward1.tolist(a) == [True, False] * 1000
    assert awkward1.tolist(tmp) == [True, False] * 100

def test_integer():
    a = awkward1.layout.FillableArray()
    a.integer(10)
    a.integer(9)
    a.integer(8)
    a.integer(7)
    a.integer(6)
    assert awkward1.tolist(a.snapshot()) == [10, 9, 8, 7, 6]
    assert awkward1.tolist(a) == [10, 9, 8, 7, 6]
    assert awkward1.tolist(a[1:-1]) == [9, 8, 7]

def test_real():
    a = awkward1.layout.FillableArray()
    a.real(1.1)
    a.real(2.2)
    a.real(3.3)
    a.real(4.4)
    a.real(5.5)
    assert awkward1.tolist(a.snapshot()) == [1.1, 2.2, 3.3, 4.4, 5.5]
    assert awkward1.tolist(a) == [1.1, 2.2, 3.3, 4.4, 5.5]
    assert awkward1.tolist(a[1:-1]) == [2.2, 3.3, 4.4]

def test_integer_real():
    a = awkward1.layout.FillableArray()
    a.integer(1)
    a.integer(2)
    a.real(3.3)
    a.integer(4)
    a.integer(5)
    assert awkward1.tolist(a.snapshot()) == [1.0, 2.0, 3.3, 4.0, 5.0]
    assert awkward1.tolist(a) == [1.0, 2.0, 3.3, 4.0, 5.0]
    assert awkward1.tolist(a[1:-1]) == [2.0, 3.3, 4.0]

def test_real_integer():
    a = awkward1.layout.FillableArray()
    a.real(1.1)
    a.real(2.2)
    a.integer(3)
    a.real(4.4)
    a.real(5.5)
    assert awkward1.tolist(a.snapshot()) == [1.1, 2.2, 3.0, 4.4, 5.5]
    assert awkward1.tolist(a) == [1.1, 2.2, 3.0, 4.4, 5.5]
    assert awkward1.tolist(a[1:-1]) == [2.2, 3.0, 4.4]

def test_list_real():
    a = awkward1.layout.FillableArray()
    a.beginlist()
    a.real(1.1)
    a.real(2.2)
    a.real(3.3)
    a.endlist()
    a.beginlist()
    a.endlist()
    a.beginlist()
    a.real(4.4)
    a.real(5.5)
    a.endlist()
    assert awkward1.tolist(a.snapshot()) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
    assert awkward1.tolist(a) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
    assert awkward1.tolist(a[1:-1]) == [[]]
    assert awkward1.tolist(a[1:]) == [[], [4.4, 5.5]]

def test_list_list_real():
    a = awkward1.layout.FillableArray()
    a.beginlist()
    a.beginlist()
    a.real(1.1)
    a.real(2.2)
    a.real(3.3)
    a.endlist()
    a.beginlist()
    a.endlist()
    a.beginlist()
    a.real(4.4)
    a.real(5.5)
    a.endlist()
    a.endlist()
    a.beginlist()
    a.endlist()
    a.beginlist()
    a.beginlist()
    a.real(6.6)
    a.real(7.7)
    a.endlist()
    a.beginlist()
    a.real(8.8)
    a.real(9.9)
    a.endlist()
    a.endlist()
    assert awkward1.tolist(a.snapshot()) == [[[1.1, 2.2, 3.3], [], [4.4, 5.5]], [], [[6.6, 7.7], [8.8, 9.9]]]
    assert awkward1.tolist(a) == [[[1.1, 2.2, 3.3], [], [4.4, 5.5]], [], [[6.6, 7.7], [8.8, 9.9]]]
    assert awkward1.tolist(a[1:]) == [[], [[6.6, 7.7], [8.8, 9.9]]]

def test_list_errors():
    with pytest.raises(ValueError):
        a = awkward1.layout.FillableArray()
        a.endlist()

    with pytest.raises(ValueError):
        a = awkward1.layout.FillableArray()
        a.real(3.14)
        a.endlist()

    with pytest.raises(ValueError):
        a = awkward1.layout.FillableArray()
        a.beginlist()
        a.real(3.14)
        a.endlist()
        a.endlist()

    with pytest.raises(ValueError):
        a = awkward1.layout.FillableArray()
        a.beginlist()
        a.beginlist()
        a.real(3.14)
        a.endlist()
        a.endlist()
        a.endlist()

    a = awkward1.layout.FillableArray()
    a.beginlist()
    a.real(1.1)
    a.real(2.2)
    a.real(3.3)
    a.endlist()
    a.beginlist()
    a.real(4.4)
    a.real(5.5)
    assert awkward1.tolist(a.snapshot()) == [[1.1, 2.2, 3.3]]
    assert awkward1.tolist(a) == [[1.1, 2.2, 3.3]]
    assert awkward1.tolist(a[1:]) == []
